(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/sandcats.js                                                  //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
var _this = this;                                                      //
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
Sandcats = {};                                                         // 17
                                                                       //
var querystring = Npm.require('querystring');                          // 19
var https = Npm.require('https');                                      // 20
var fs = Npm.require('fs');                                            // 21
var dgram = Npm.require('dgram');                                      // 22
var Url = Npm.require('url');                                          // 23
                                                                       //
var SANDCATS_HOSTNAME = Meteor.settings && Meteor.settings['public'] && Meteor.settings['public'].sandcatsHostname;
var SANDCATS_VARDIR = (SANDSTORM_ALTHOME || '') + '/var/sandcats';     // 27
                                                                       //
var ROOT_URL = Url.parse(process.env.ROOT_URL);                        // 29
var HOSTNAME = ROOT_URL.hostname;                                      // 30
var SANDCATS_NAME = undefined; // Look at `startup` below to see where this is set
                                                                       //
var updateSandcatsIp = function () {                                   // 33
  var responseCallback = function (res) {                              // 34
    if (res.statusCode === 200) {                                      // 35
      console.log('Successfully updated sandcats IP');                 // 36
    } else {                                                           //
      console.error('Failed to update sandcats IP:', res.headers);     // 38
    }                                                                  //
  };                                                                   //
                                                                       //
  var errorCallback = function (err) {                                 // 42
    console.error('Couldn\'t send update sandcats hostname', err);     // 43
  };                                                                   //
                                                                       //
  performSandcatsRequest('/update', SANDCATS_HOSTNAME, { rawHostname: SANDCATS_NAME }, errorCallback, responseCallback);
};                                                                     //
                                                                       //
var pingUdp = function () {                                            // 50
  var socket = dgram.createSocket('udp4');                             // 51
  var secret = Random.secret(16);                                      // 52
                                                                       //
  message = new Buffer(SANDCATS_NAME + ' ' + secret);                  // 54
  socket.on('message', function (buf) {                                // 55
    if (buf.toString() === secret) {                                   // 56
      updateSandcatsIp();                                              // 57
    } else {                                                           //
      console.error('Received unexpected response in UDP sandcats ping:', buf.toString());
    }                                                                  //
  });                                                                  //
                                                                       //
  socket.send(message, 0, message.length, 8080, SANDCATS_HOSTNAME, function (err) {
    if (err) {                                                         // 64
      console.error('Couldn\'t send UDP sandcats ping', err);          // 65
    }                                                                  //
  });                                                                  //
                                                                       //
  Meteor.setTimeout(function () {                                      // 69
    socket.close();                                                    // 70
  }, 10 * 1000);                                                       //
};                                                                     //
                                                                       //
var performSandcatsRequest = function (path, hostname, postData, errorCallback, responseCallback) {
  var options = {                                                      // 75
    hostname: hostname,                                                // 76
    path: path,                                                        // 77
    method: 'POST',                                                    // 78
    agent: false,                                                      // 79
    key: fs.readFileSync(SANDCATS_VARDIR + '/id_rsa'),                 // 80
    cert: fs.readFileSync(SANDCATS_VARDIR + '/id_rsa.pub'),            // 81
    headers: {                                                         // 82
      'X-Sand': 'cats',                                                // 83
      'Content-Type': 'application/x-www-form-urlencoded'              // 84
    }                                                                  //
  };                                                                   //
                                                                       //
  if (postData.certificateSigningRequest) {                            // 88
    console.log('Submitting certificate request for host', postData.rawHostname, 'where the request has length', postData.certificateSigningRequest.length);
  }                                                                    //
                                                                       //
  var newPostData = querystring.stringify(postData);                   // 94
  if (SANDCATS_HOSTNAME === 'sandcats-dev.sandstorm.io' && !process.env.NODE_TLS_REJECT_UNAUTHORIZED) {
    console.log('You are using the Sandcats dev server but have full HTTPS ' + 'certificate checking enabled. Probably you will get an HTTPS ' + 'error. Proceeding anyway.');
  }                                                                    //
                                                                       //
  var req = https.request(options, responseCallback);                  // 102
  req.write(newPostData);                                              // 103
  req.end();                                                           // 104
                                                                       //
  if (errorCallback) {                                                 // 106
    req.on('error', errorCallback);                                    // 107
  }                                                                    //
                                                                       //
  return req;                                                          // 110
};                                                                     //
                                                                       //
var generateKeyAndCsr = function (commonName) {                        // 113
  // This function relies on the this.forge object created by the      //
  // meteor-node-forge package.                                        //
  check(commonName, String);                                           // 116
                                                                       //
  // Generate key pair. Using Meteor.wrapAsync because forge supports  //
  // a synchronous as well as an asynchronous API, and the synchronous
  // one blocks for a while.                                           //
  var wrappedGenerateKeyPair = Meteor.wrapAsync(_this.forge.pki.rsa.generateKeyPair);
                                                                       //
  // I could pick an `e`[xponent] value for the resulting RSA key, but
  // I will refrain.                                                   //
  var keys = wrappedGenerateKeyPair({ bits: 2048 });                   // 125
                                                                       //
  // Create a certificate request (CSR).                               //
  var csr = _this.forge.pki.createCertificationRequest();              // 128
  csr.publicKey = keys.publicKey;                                      // 129
  csr.setSubject([{                                                    // 130
    name: 'commonName',                                                // 132
    value: commonName                                                  // 133
  }]);                                                                 //
  csr.sign(keys.privateKey);                                           // 136
  console.log('generateKeyAndCsr created new key & certificate request for', commonName);
  return {                                                             // 138
    privateKeyAsPem: _this.forge.pki.privateKeyToPem(keys.privateKey),
    csrAsPem: _this.forge.pki.certificationRequestToPem(csr)           // 140
  };                                                                   //
};                                                                     //
                                                                       //
Sandcats.storeNewKeyAndCsr = function (hostname, basePath) {           // 144
  // We use the current JS time (like UNIX timestamp but in            //
  // milliseconds) as the key number. Note that the keyNumber is       //
  // intended to be an opaque identifier; the only important thing is  //
  // that it increases numerically over time. We use the current time  //
  // just as a simplistic way to pick a filename that probably no one  //
  // else has created yet.                                             //
  var keyNumber = new Date().getTime();                                // 151
  var keyFilename = basePath + '/' + keyNumber;                        // 152
  var csrFilename = keyFilename + '.csr';                              // 153
  var responseFilename = keyFilename + '.response-json';               // 154
  var withWildcard = '*.' + hostname;                                  // 155
  var keyAndCsr = generateKeyAndCsr(withWildcard);                     // 156
  fs.writeFileSync(keyFilename, keyAndCsr.privateKeyAsPem, { mode: 0400 });
  fs.writeFileSync(csrFilename, keyAndCsr.csrAsPem);                   // 159
  console.log('storeNewKeyAndCsr successfully saved key and certificate request to', keyFilename, 'and', csrFilename, 'respectively of length', keyAndCsr.privateKeyAsPem.length, 'and', keyAndCsr.csrAsPem.length);
  return {                                                             // 164
    csrFilename: csrFilename,                                          // 165
    keyFilename: keyFilename,                                          // 166
    responseFilename: responseFilename                                 // 167
  };                                                                   //
};                                                                     //
                                                                       //
Sandcats.renewHttpsCertificateIfNeeded = function () {                 // 171
  var renewHttpsCertificate = function () {                            // 172
    var hostname = Url.parse(process.env.ROOT_URL).hostname;           // 173
    var basePath = '/var/sandcats/https/' + hostname;                  // 174
    var filenames = Sandcats.storeNewKeyAndCsr(hostname, basePath);    // 175
                                                                       //
    var errorCallback = function (err) {                               // 177
      console.error('Error while renewing HTTPS certificate (will continue to retry)', err);
    };                                                                 //
                                                                       //
    var responseCallback = function (res) {                            // 181
      if (res.statusCode == 200) {                                     // 182
        (function () {                                                 //
          // Save the response, chunk by chunk, then store it on disk  //
          // for later use.                                            //
          var responseBody = '';                                       // 185
          res.on('data', function (chunk) {                            // 186
            responseBody += chunk;                                     // 187
          });                                                          //
                                                                       //
          res.on('end', function () {                                  // 190
            // For sanity, make sure it parses as JSON, since we're    //
            // going to need it to do that.                            //
            try {                                                      // 193
              JSON.parse(responseBody);                                // 194
            } catch (e) {                                              //
              console.error('JSON parse error receiving new HTTPS certificate. Discarding:', responseBody, 'due to exception:', e);
              // Overwrite the responseBody with the empty JSON object
              // and continue with the process of saving it to disk.   //
              responseBody = '{}';                                     // 202
            }                                                          //
                                                                       //
            try {                                                      // 205
              fs.writeFileSync(filenames.responseFilename, responseBody, 'utf-8');
            } catch (err) {                                            //
              return console.error('Failure while saveing new HTTPS certificate to', filenames.responseFilename, 'will continue to retry. Exception was:', err);
            }                                                          //
                                                                       //
            // Tell Sandcats that now is a good time to update its info
            // about which keys are available.                         //
            if (!global.sandcats) {                                    // 216
              console.error('When getting new certificate, could not find callback to request re-keying! Certs will probably become invalid soon.');
            } else {                                                   //
              // Call the sandcats rekeying function.                  //
              global.sandcats.rekey();                                 // 220
              // That's that.                                          //
              console.log('Successfully renewed HTTPS certificate into', filenames.responseFilename);
            }                                                          //
          });                                                          //
        })();                                                          //
      } else {                                                         //
        console.log('Received HTTP error while renewing certificate (will keep retrying)', res.statusCode);
        res.on('data', function (chunk) {                              // 229
          console.log('Error response contained information', chunk.toString('utf-8'));
        });                                                            //
      }                                                                //
    };                                                                 //
                                                                       //
    var wrappedReadFile = Meteor.wrapAsync(fs.readFile);               // 235
                                                                       //
    performSandcatsRequest('/getcertificate', SANDCATS_HOSTNAME, {     // 237
      rawHostname: SANDCATS_NAME,                                      // 238
      certificateSigningRequest: wrappedReadFile(filenames.csrFilename, 'utf-8')
    }, errorCallback, responseCallback);                               //
  };                                                                   //
                                                                       //
  if (global.sandcats.shouldGetAnotherCertificate()) {                 // 243
    console.log('renewHttpsCertificateIfNeeded: Happily choosing to renew certificate.');
    return renewHttpsCertificate();                                    // 245
  }                                                                    //
};                                                                     //
                                                                       //
Sandcats.initializeSandcats = function () {                            // 249
  var i = HOSTNAME.lastIndexOf(SANDCATS_HOSTNAME);                     // 250
  if (i < 0) {                                                         // 251
    console.error('SANDCATS_BASE_DOMAIN is configured but your HOSTNAME doesn\'t appear to contain it:', SANDCATS_HOSTNAME, HOSTNAME);
  } else {                                                             //
    var oneMinute = 60 * 1000;                                         // 255
    var oneHour = 60 * oneMinute;                                      // 256
    var randomIntervalZeroToOneHour = Math.random() * oneHour;         // 257
    // All Sandcats installs need dyndns updating.                     //
    SANDCATS_NAME = HOSTNAME.slice(0, i - 1);                          // 259
    Meteor.setInterval(pingUdp, oneMinute);                            // 260
    // If process.env.HTTPS_PORT is set, we need to auto-refresh our HTTPS certificate.
    if (process.env.HTTPS_PORT) {                                      // 262
      // Always do a HTTPS certificate update check on Sandstorm start.
      Sandcats.renewHttpsCertificateIfNeeded();                        // 264
                                                                       //
      // After that's done, schedule it for every approx 1-2 hours in  //
      // the future.                                                   //
      Meteor.setInterval(Sandcats.renewHttpsCertificateIfNeeded, oneHour + randomIntervalZeroToOneHour);
    }                                                                  //
  }                                                                    //
};                                                                     //
                                                                       //
if (SANDCATS_HOSTNAME) {                                               // 274
  Meteor.startup(Sandcats.initializeSandcats);                         // 275
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=sandcats.js.map
