(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/grain.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file implements /grain, i.e. the main view into an app.        //
                                                                       //
var DEFAULT_TITLE = "Sandstorm";                                       // 19
                                                                       //
if (Meteor.isServer) {                                                 // 21
  var Crypto = Npm.require("crypto");                                  // 22
                                                                       //
  Meteor.publish("grainTopBar", function (grainId) {                   // 24
    check(grainId, String);                                            // 25
    var self = this;                                                   // 26
                                                                       //
    var result = [Grains.find({ _id: grainId, $or: [{ userId: this.userId }, { "private": { $ne: true } }] }, { fields: { title: 1, userId: 1, identityId: 1, "private": 1 } })];
    if (this.userId) {                                                 // 30
      var myIdentityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
      myIdentityIds.forEach(function (id) {                            // 32
        var identity = globalDb.getUser(id);                           // 33
        self.added("displayNames", identity._id, { displayName: identity.profile.name });
      });                                                              //
                                                                       //
      // Alice is allowed to know Bob's display name if Bob has received a UiView from Alice
      // for *any* grain.                                              //
      var handle = ApiTokens.find({ identityId: { $in: myIdentityIds },
        "owner.user.identityId": { $exists: true } }).observe({        // 40
        added: function (token) {                                      // 41
          var user = Meteor.users.findOne({ _id: token.owner.user.identityId });
          if (user) {                                                  // 43
            var identity = globalDb.getUser(token.owner.user.identityId);
            if (identity) {                                            // 45
              self.added("displayNames", identity._id, { displayName: identity.profile.name });
            }                                                          //
          }                                                            //
        }                                                              //
      });                                                              //
      this.onStop(function () {                                        // 52
        handle.stop();                                                 // 52
      });                                                              //
                                                                       //
      result.push(ApiTokens.find({ grainId: grainId,                   // 54
        $or: [{ "owner.user.identityId": { $in: myIdentityIds } }, { identityId: { $in: myIdentityIds } }] }));
    }                                                                  //
    return result;                                                     // 59
  });                                                                  //
                                                                       //
  // We allow users to learn package information about a grain they own.
  // This is used for obtaining icon and app title information for grains
  // you own, which is used in the sidebar. It is not a security/privacy
  // risk since it only exposes this information for grains the user owns.
  Meteor.publish("packageByGrainId", function (grainId) {              // 66
    check(grainId, String);                                            // 67
    var publishThis = [];                                              // 68
    // We need to publish the packageId so that client-side code can   //
    // find the right package.                                         //
    var thisGrainCursor = Grains.find({ _id: grainId, userId: this.userId }, { fields: { packageId: 1 } });
    publishThis.push(thisGrainCursor);                                 // 73
                                                                       //
    if (thisGrainCursor.count()) {                                     // 75
      var thisGrain = thisGrainCursor.fetch()[0];                      // 76
      var thisPackageCursor = Packages.find({ _id: thisGrain.packageId });
      publishThis.push(thisPackageCursor);                             // 78
    }                                                                  //
                                                                       //
    return publishThis;                                                // 81
  });                                                                  //
                                                                       //
  Meteor.publish("tokenInfo", function (token) {                       // 84
    // Allows the client side to map a raw token to its entry in ApiTokens, and the additional
    // metadata that it will need to display the app icon and title.  We do not care about making
    // the metadata reactive.                                          //
    check(token, String);                                              // 88
                                                                       //
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
    var apiToken = ApiTokens.findOne({ _id: hashedToken }, { fields: { grainId: 1, owner: 1 } });
    if (!apiToken) {                                                   // 93
      this.added("tokenInfo", token, { invalidToken: true });          // 94
    } else {                                                           //
      var grainId = apiToken.grainId;                                  // 96
      var grain = Grains.findOne({ _id: grainId }, { fields: { packageId: 1, appId: 1 } });
      if (!grain) {                                                    // 98
        this.added("tokenInfo", token, { invalidToken: true });        // 99
      } else {                                                         //
        if (apiToken.owner && apiToken.owner.user) {                   // 101
          var identity = globalDb.getIdentity(apiToken.owner.user.identityId);
          var metadata = apiToken.owner.user.denormalizedGrainMetadata;
          if (identity && metadata) {                                  // 104
            this.added("tokenInfo", token, { identityOwner: _.pick(identity, "_id", "profile"),
              grainId: grainId,                                        // 107
              grainMetadata: metadata });                              // 108
          } else {                                                     //
            this.added("tokenInfo", token, { invalidToken: true });    // 110
          }                                                            //
        } else if (!apiToken.owner || "webkey" in apiToken.owner) {    //
          var pkg = Packages.findOne({ _id: grain.packageId }, { fields: { manifest: 1 } });
          var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };
          var appIcon = undefined;                                     // 115
          if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {
            var icons = pkg.manifest.metadata.icons;                   // 117
            appIcon = icons.grain || icons.appGrid;                    // 118
          }                                                            //
          var denormalizedGrainMetadata = {                            // 120
            appTitle: appTitle,                                        // 121
            icon: appIcon,                                             // 122
            appId: appIcon ? undefined : grain.appId                   // 123
          };                                                           //
          this.added("tokenInfo", token, { webkey: true,               // 125
            grainId: grainId,                                          // 127
            grainMetadata: denormalizedGrainMetadata });               // 128
        } else {                                                       //
          this.added("tokenInfo", token, { invalidToken: true });      // 130
        }                                                              //
      }                                                                //
    }                                                                  //
    this.ready();                                                      // 134
    return;                                                            // 135
  });                                                                  //
                                                                       //
  Meteor.publish("grainSize", function (grainId) {                     // 138
    // Publish pseudo-collection containing the size of the grain opened in the given session.
    check(grainId, String);                                            // 140
                                                                       //
    var grain = Grains.findOne(grainId);                               // 142
    if (!grain || grain.userId !== this.userId) {                      // 143
      return [];                                                       // 144
    }                                                                  //
                                                                       //
    var supervisor = globalBackend.cap().getGrain(this.userId, grainId).supervisor;
                                                                       //
    var self = this;                                                   // 149
    var stopped = false;                                               // 150
    var promise = getGrainSize(supervisor);                            // 151
                                                                       //
    function getNext(oldSize) {                                        // 153
      promise = getGrainSize(supervisor, oldSize);                     // 154
      promise.then(function (size) {                                   // 155
        if (!stopped) {                                                // 156
          if (size !== oldSize) {                                      // 157
            // sometimes there are false alarms                        //
            self.changed("grainSizes", grainId, { size: size });       // 158
          }                                                            //
          getNext(size);                                               // 160
        }                                                              //
      }, function (err) {                                              //
        if (!stopped) {                                                // 163
          if (err.kjType === "disconnected") {                         // 164
            self.stop();                                               // 165
          } else {                                                     //
            self.error(err);                                           // 167
          }                                                            //
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    promise.then(function (size) {                                     // 173
      if (!stopped) {                                                  // 174
        self.added("grainSizes", grainId, { size: size });             // 175
        self.ready();                                                  // 176
        getNext(size);                                                 // 177
      }                                                                //
    }, function (err) {                                                //
      if (!stopped) {                                                  // 180
        if (err.kjType === "disconnected") {                           // 181
          self.stop();                                                 // 182
        } else {                                                       //
          self.error(err);                                             // 184
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    self.onStop(function () {                                          // 189
      stopped = true;                                                  // 190
      promise.cancel();                                                // 191
    });                                                                //
  });                                                                  //
}                                                                      //
                                                                       //
// GrainSizes is used by grainview.js                                  //
GrainSizes = new Mongo.Collection("grainSizes");                       // 197
var DisplayNames = new Mongo.Collection("displayNames");               // 198
// TokenInfo is used by grainview.js                                   //
TokenInfo = new Mongo.Collection("tokenInfo");                         // 200
// Pseudo-collections published above.                                 //
                                                                       //
Meteor.methods({                                                       // 203
  deleteGrain: function (grainId) {                                    // 204
    check(grainId, String);                                            // 205
                                                                       //
    if (this.userId) {                                                 // 207
      var grain = Grains.findOne({ _id: grainId, userId: this.userId });
      if (grain) {                                                     // 209
        Grains.remove(grainId);                                        // 210
        ApiTokens.remove({ grainId: grainId, $or: [{ owner: { $exists: false } }, { owner: { webkey: null } }] });
        if (grain.lastUsed) {                                          // 213
          DeleteStats.insert({ type: isDemoUser() ? "demoGrain" : "grain",
            lastActive: grain.lastUsed, appId: grain.appId });         // 215
        }                                                              //
        if (!this.isSimulation) {                                      // 217
          waitPromise(globalBackend.deleteGrain(grainId, this.userId));
          Meteor.call("deleteUnusedPackages", grain.appId);            // 219
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
  forgetGrain: function (grainId, identityId) {                        // 224
    // TODO(cleanup): For now we are ignoring `identityId`, but maybe we should expose finer-grained
    //  forgetting.                                                    //
                                                                       //
    check(grainId, String);                                            // 228
    check(identityId, String);                                         // 229
                                                                       //
    if (!this.userId) {                                                // 231
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");
    }                                                                  //
    SandstormDb.getUserIdentityIds(Meteor.user()).forEach(function (identityId) {
      ApiTokens.remove({ grainId: grainId, "owner.user.identityId": identityId });
    });                                                                //
  },                                                                   //
  updateGrainPreferredIdentity: function (grainId, identityId) {       // 238
    check(grainId, String);                                            // 239
    check(identityId, String);                                         // 240
    if (!this.userId) {                                                // 241
      throw new Meteor.Error(403, "Must be logged in.");               // 242
    }                                                                  //
    var grain = globalDb.getGrain(grainId) || {};                      // 244
    if (!grain.userId === this.userId) {                               // 245
      throw new Meteor.Error(403, "Grain not owned by current user.");
    }                                                                  //
    Grains.update({ _id: grainId }, { $set: { identityId: identityId } });
  },                                                                   //
  updateGrainTitle: function (grainId, newTitle, identityId) {         // 250
    check(grainId, String);                                            // 251
    check(newTitle, String);                                           // 252
    check(identityId, String);                                         // 253
    if (this.userId) {                                                 // 254
      var grain = Grains.findOne(grainId);                             // 255
      if (grain) {                                                     // 256
        if (grain.userId === this.userId) {                            // 257
          Grains.update({ _id: grainId, userId: this.userId }, { $set: { title: newTitle } });
        } else {                                                       //
          if (!globalDb.userHasIdentity(this.userId, identityId)) {    // 260
            throw new Meteor.Error(403, "Current user does not have identity " + identityId);
          }                                                            //
          var token = ApiTokens.findOne({ grainId: grainId, objectId: { $exists: false },
            "owner.user.identityId": identityId }, { sort: { created: 1 } });
          if (token) {                                                 // 266
            ApiTokens.update(token._id, { $set: { "owner.user.title": newTitle } });
          }                                                            //
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
  privatizeGrain: function (grainId) {                                 // 273
    check(grainId, String);                                            // 274
    if (this.userId) {                                                 // 275
      Grains.update({ _id: grainId, userId: this.userId }, { $set: { "private": true } });
    }                                                                  //
  },                                                                   //
  inviteUsersToGrain: function (origin, identityId, grainId, title, roleAssignment, contacts, message) {
    if (!this.isSimulation) {                                          // 281
      check(origin, String);                                           // 282
      check(identityId, String);                                       // 283
      check(grainId, String);                                          // 284
      check(title, String);                                            // 285
      check(roleAssignment, roleAssignmentPattern);                    // 286
      check(contacts, [{                                               // 287
        _id: String,                                                   // 288
        isDefault: Match.Optional(Boolean),                            // 289
        profile: Match.ObjectIncluding({                               // 290
          service: String,                                             // 291
          name: String,                                                // 292
          intrinsicName: String                                        // 293
        }) }]);                                                        //
      check(message, { text: String, html: String });                  // 295
      if (!this.userId) {                                              // 296
        throw new Meteor.Error(403, "Must be logged in to share by email.");
      }                                                                //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {        // 299
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
      }                                                                //
      if (contacts.length === 0) {                                     // 302
        throw new Meteor.Error(400, "No contacts were provided.");     // 303
      }                                                                //
      var accountId = this.userId;                                     // 305
      var identity = globalDb.getIdentity(identityId);                 // 306
      var sharerDisplayName = identity.profile.name;                   // 307
      var outerResult = { successes: [], failures: [] };               // 308
      contacts.forEach(function (contact) {                            // 309
        if (contact.isDefault && contact.profile.service === "email") {
          var emailAddress = contact.profile.intrinsicName;            // 311
          var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "email invitation for " + emailAddress, roleAssignment, { webkey: { forSharing: true } });
          var url = origin + "/shared/" + result.token;                // 316
          var html = message.html + "<br><br>" + "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + "Open Shared Grain</a><div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
          try {                                                        // 324
            SandstormEmail.send({                                      // 325
              to: emailAddress,                                        // 326
              from: "Sandstorm server <no-reply@" + HOSTNAME + ">",    // 327
              subject: sharerDisplayName + " has invited you to join a grain: " + title,
              text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
              html: html                                               // 332
            });                                                        //
          } catch (e) {                                                //
            outerResult.failures.push({ contact: contact, error: e.toString() });
          }                                                            //
        } else {                                                       //
          var _result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "direct invitation to " + contact.profile.intrinsicName, roleAssignment, { user: { identityId: contact._id, title: title } });
          try {                                                        // 342
            var identity = Meteor.users.findOne({ _id: contact._id });
            var emailAddress = SandstormDb.getVerifiedEmails(identity)[0];
            var url = origin + "/shared/" + _result.token;             // 345
            if (emailAddress) {                                        // 346
              var intrinsicName = contact.profile.intrinsicName;       // 347
              var loginNote;                                           // 348
              if (contact.profile.service === "google") {              // 349
                loginNote = "Google account with address " + emailAddress;
              } else if (contact.profile.service === "github") {       //
                loginNote = "Github account with username " + intrinsicName;
              } else if (contact.profile.service === "email") {        //
                loginNote = "email address " + intrinsicName;          // 354
              } else {                                                 //
                throw new Meteor.Error(500, "Unknown service to email share link.");
              }                                                        //
              var html = message.html + "<br><br>" + "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + "Open Shared Grain</a><div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: You will need to log in with your " + loginNote + " to access this grain.";
              SandstormEmail.send({                                    // 365
                to: emailAddress,                                      // 366
                from: "Sandstorm server <no-reply@" + HOSTNAME + ">",  // 367
                subject: sharerDisplayName + " has invited you to join a grain: " + title,
                text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: You will need to log in with your " + loginNote + " to access this grain.",
                html: html                                             // 372
              });                                                      //
            } else {                                                   //
              outerResult.failures.push({ contact: contact, warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + url + " with them." });
            }                                                          //
          } catch (e) {                                                //
            outerResult.failures.push({ contact: contact, error: e.toString(),
              warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + url + " with them." });
          }                                                            //
        }                                                              //
      });                                                              //
      return outerResult;                                              // 386
    }                                                                  //
  }                                                                    //
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 391
  Tracker.autorun(function () {                                        // 392
    // We need to keep track of certain data about each grain we can view.
    // TODO(cleanup): Do these in GrainView to avoid spurious resubscribes.
    var grains = globalGrains.get();                                   // 395
    grains.forEach(function (grain) {                                  // 396
      grain.depend();                                                  // 397
      var grainId = grain.grainId();                                   // 398
      if (grainId) {                                                   // 399
        Meteor.subscribe("grainTopBar", grainId);                      // 400
        if (grain.isOwner()) {                                         // 401
          Meteor.subscribe("packageByGrainId", grainId);               // 402
        }                                                              //
        var token = grain.token();                                     // 404
        if (token) {                                                   // 405
          Meteor.subscribe("tokenInfo", token);                        // 406
        }                                                              //
      }                                                                //
    });                                                                //
  });                                                                  //
                                                                       //
  Template.layout.events({                                             // 412
    "click .incognito-button": function (event) {                      // 413
      console.log(event);                                              // 414
      var grains = globalGrains.get();                                 // 415
      var token = event.currentTarget.getAttribute("data-token");      // 416
      if (token) {                                                     // 417
        grains.forEach(function (grain) {                              // 418
          if (grain.token() == token) {                                // 419
            grain.doNotRevealIdentity();                               // 420
          }                                                            //
        });                                                            //
      } else {                                                         //
        console.error("Interstitial prompt answered, but no token present?");
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  var promptNewTitle = function () {                                   // 429
    var grain = getActiveGrain(globalGrains.get());                    // 430
    if (grain) {                                                       // 431
      var prompt = "Set new title:";                                   // 432
      if (!grain.isOwner()) {                                          // 433
        prompt = "Set a new personal title: (does not change the owner's title for this grain)";
      }                                                                //
      var title = window.prompt(prompt, grain.title());                // 436
      if (title) {                                                     // 437
        grain.setTitle(title);                                         // 438
      }                                                                //
    }                                                                  //
  };                                                                   //
  Template.grainTitle.events({                                         // 442
    "click": function (event) {                                        // 443
      promptNewTitle();                                                // 444
    },                                                                 //
    "keydown": function (event) {                                      // 446
      if (event.keyCode === 13 || event.keyCode === 32) {              // 447
        // Allow space or enter to trigger renaming the grain - Firefox doesn't treat enter on the
        // focused element as click().                                 //
        promptNewTitle();                                              // 450
        event.preventDefault();                                        // 451
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainDeleteButton.events({                                  // 456
    "click button": function (event) {                                 // 457
      var grains = globalGrains.get();                                 // 458
      var activeIndex = activeGrainIndex(grains);                      // 459
      var activeGrain = grains[activeIndex];                           // 460
      var newActiveIndex = activeIndex == grains.length - 1 ? activeIndex - 1 : activeIndex;
      if (activeGrain.isOwner()) {                                     // 462
        if (window.confirm("Really delete this grain?")) {             // 463
          Meteor.call("deleteGrain", activeGrain.grainId());           // 464
          // TODO: extract globalGrains into a class that has a "close" method for closing the active view
          activeGrain.destroy();                                       // 466
          if (grains.length == 1) {                                    // 467
            globalGrains.set([]);                                      // 468
            Router.go("root"); // will redirect to the right place     // 469
          } else {                                                     //
              grains.splice(activeIndex, 1);                           // 471
              grains[newActiveIndex].setActive(true);                  // 472
              globalGrains.set(grains);                                // 473
              Router.go(grains[newActiveIndex].route());               // 474
            }                                                          //
        }                                                              //
      } else {                                                         //
        if (window.confirm("Really forget this grain?")) {             // 478
          var identityId = activeGrain.identityId();                   // 479
          if (identityId) {                                            // 480
            Meteor.call("forgetGrain", activeGrain.grainId(), identityId);
          }                                                            //
          // TODO: extract globalGrains into a class that has a "close" method for closing the active view
          activeGrain.destroy();                                       // 484
          if (grains.length == 1) {                                    // 485
            globalGrains.set([]);                                      // 486
            Router.go("root"); // will redirect to the right place     // 487
          } else {                                                     //
              grains.splice(activeIndex, 1);                           // 489
              grains[newActiveIndex].setActive(true);                  // 490
              globalGrains.set(grains);                                // 491
              Router.go(grains[newActiveIndex].route());               // 492
            }                                                          //
        }                                                              //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainDebugLogButton.events({                                // 499
    "click button": function (event) {                                 // 500
      this.reset();                                                    // 501
      var activeGrain = getActiveGrain(globalGrains.get());            // 502
      window.open("/grainlog/" + activeGrain.grainId(), "_blank", "menubar=no,status=no,toolbar=no,width=700,height=700");
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainBackupButton.onCreated(function () {                   // 508
    this.isLoading = new ReactiveVar(false);                           // 509
  });                                                                  //
                                                                       //
  Template.grainBackupButton.helpers({                                 // 512
    isLoading: function () {                                           // 513
      return Template.instance().isLoading.get();                      // 514
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainBackupButton.events({                                  // 518
    "click button": function (event, template) {                       // 519
      template.isLoading.set(true);                                    // 520
      this.reset();                                                    // 521
      var activeGrain = getActiveGrain(globalGrains.get());            // 522
      Meteor.call("backupGrain", activeGrain.grainId(), function (err, id) {
        template.isLoading.set(false);                                 // 524
        if (err) {                                                     // 525
          alert("Backup failed: " + err); // TODO(someday): make this better UI
        } else {                                                       //
            // Firefox for some reason decides to kill all websockets when we try to download the file
            // by navigating there. So we're left doing a dirty hack to get around the popup blocker.
            var isFirefox = typeof InstallTrigger !== "undefined";     // 530
                                                                       //
            if (isFirefox) {                                           // 532
              var save = document.createElement("a");                  // 533
              save.href = "/downloadBackup/" + id;                     // 534
                                                                       //
              save.download = activeGrain.title() + ".zip";            // 536
              var event = document.createEvent("MouseEvents");         // 537
              event.initMouseEvent("click", true, false, window, 0, 0, 0, 0, 0, false, false, false, false, 0, null);
              save.dispatchEvent(event);                               // 542
            } else {                                                   //
              window.location = "/downloadBackup/" + id;               // 544
            }                                                          //
          }                                                            //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainRestartButton.events({                                 // 551
    "click button": function (event) {                                 // 552
      this.reset();                                                    // 553
      var activeGrain = getActiveGrain(globalGrains.get());            // 554
      var grainId = activeGrain.grainId();                             // 555
                                                                       //
      Meteor.call("shutdownGrain", grainId, function (err) {           // 557
        if (err) {                                                     // 558
          alert("Restart failed: " + err); // TODO(someday): make this better UI
        } else {                                                       //
            var frames = document.getElementsByClassName("grain-frame");
            for (var i = 0; i < frames.length; i++) {                  // 562
              var frame = frames[i];                                   // 563
              if (frame.dataset.grainid == grainId) {                  // 564
                frame.src = frame.src;                                 // 565
              }                                                        //
            }                                                          //
          }                                                            //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  function selectElementContents(element) {                            // 573
    if (document.body.createTextRange) {                               // 574
      var range = document.body.createTextRange();                     // 575
      range.moveToElementText(element);                                // 576
      range.select();                                                  // 577
    } else if (window.getSelection) {                                  //
      var selection = window.getSelection();                           // 579
      var range = document.createRange();                              // 580
      range.selectNodeContents(element);                               // 581
      selection.removeAllRanges();                                     // 582
      selection.addRange(range);                                       // 583
    }                                                                  //
  }                                                                    //
                                                                       //
  function selectTargetContents(event) {                               // 587
    event.preventDefault();                                            // 588
    selectElementContents(event.currentTarget);                        // 589
  }                                                                    //
                                                                       //
  Template.grainApiTokenPopup.events({                                 // 592
    "click .copy-me": selectTargetContents,                            // 593
    "focus .copy-me": selectTargetContents,                            // 594
    "submit .newApiToken": function (event) {                          // 595
      event.preventDefault();                                          // 596
      var activeGrain = getActiveGrain(globalGrains.get());            // 597
      var grainId = activeGrain.grainId();                             // 598
      activeGrain.setGeneratedApiToken("pending");                     // 599
      var roleList = document.getElementById("api-token-role");        // 600
      // TODO(cleanup): avoid using global ids; select a child of the current template instead
      var assignment = { allAccess: null };                            // 602
      if (roleList && roleList.selectedIndex > 0) {                    // 603
        assignment = { roleId: roleList.selectedIndex - 1 };           // 604
      }                                                                //
      Meteor.call("newApiToken", { identityId: activeGrain.identityId() }, grainId, document.getElementById("api-token-petname").value, assignment, { webkey: { forSharing: false } }, function (error, result) {
        if (error) {                                                   // 610
          activeGrain.setGeneratedApiToken(undefined);                 // 611
          window.alert("Failed to create token.\n" + error);           // 612
          console.error(error.stack);                                  // 613
        } else {                                                       //
          activeGrain.setGeneratedApiToken(result.endpointUrl + "#" + result.token);
        }                                                              //
      });                                                              //
    },                                                                 //
    "click #resetApiToken": function (event) {                         // 619
      var activeGrain = getActiveGrain(globalGrains.get());            // 620
      activeGrain.setGeneratedApiToken(undefined);                     // 621
    },                                                                 //
    "click button.revoke-token": function (event) {                    // 623
      Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
    },                                                                 //
                                                                       //
    "click .token-petname": function (event) {                         // 628
      // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
      //   to edit the petname in place.                               //
      var petname = window.prompt("Set new label:", this.petname);     // 631
      if (petname) {                                                   // 632
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainSharePopup.events({                                    // 639
    "click .copy-me": selectTargetContents,                            // 640
    "focus .copy-me": selectTargetContents,                            // 641
    "click #share-grain-popup-closer": function (event) {              // 642
      Session.set("show-share-grain", false);                          // 643
    },                                                                 //
    "click button.who-has-access": function (event, instance) {        // 645
      event.preventDefault();                                          // 646
      var closer = globalTopbar.addItem({                              // 647
        name: "who-has-access",                                        // 648
        template: Template.whoHasAccess,                               // 649
        popupTemplate: Template.whoHasAccessPopup,                     // 650
        data: new ReactiveVar(instance.data),                          // 651
        startOpen: true,                                               // 652
        onDismiss: function () {                                       // 653
          return "remove";                                             // 654
        }                                                              //
      });                                                              //
    },                                                                 //
    "click #privatize-grain": function (event) {                       // 658
      Meteor.call("privatizeGrain", getActiveGrain(globalGrains.get()).grainId());
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.shareWithOthers.onRendered(function () {                    // 663
    this.find("[role=tab]").focus();                                   // 664
  });                                                                  //
  var activateTargetTab = function (event, instance) {                 // 666
    // Deactivate all tabs and all tab panels.                         //
    instance.findAll("ul[role=tablist]>li[role=tab]").forEach(function (element) {
      element.setAttribute("aria-selected", false);                    // 669
    });                                                                //
    instance.findAll(".tabpanel").forEach(function (element) {         // 671
      element.setAttribute("aria-hidden", true);                       // 672
    });                                                                //
                                                                       //
    // Activate the tab header the user selected.                      //
    event.currentTarget.setAttribute("aria-selected", true);           // 676
    // Show the corresponding tab panel.                               //
    var idToShow = event.currentTarget.getAttribute("aria-controls");  // 678
    var tabPanelToShow = instance.find("#" + idToShow);                // 679
    tabPanelToShow.setAttribute("aria-hidden", false);                 // 680
  };                                                                   //
  Template.shareWithOthers.events({                                    // 682
    "click #send-invite-tab-header": activateTargetTab,                // 683
    "click #shareable-link-tab-header": activateTargetTab,             // 684
    "keydown [role=tab]": function (event, template) {                 // 685
      if (event.keyCode == 38 || event.keyCode == 40) {                // 686
        // up and down arrows                                          //
        event.preventDefault();                                        // 687
      }                                                                //
      var focus = $(template.find(":focus"));                          // 689
      var items = template.$("[role=tab]:visible");                    // 690
      var focusIndex = items.index(focus);                             // 691
      var newFocusIndex;                                               // 692
      if (event.keyCode == 37) {                                       // 693
        // left arrow                                                  //
        event.preventDefault();                                        // 694
        newFocusIndex = focusIndex - 1;                                // 695
        if (newFocusIndex == -1) {                                     // 696
          newFocusIndex = items.length - 1;                            // 697
        }                                                              //
      } else if (event.keyCode == 39) {                                //
        // right arrow                                                 //
        event.preventDefault();                                        // 700
        newFocusIndex = focusIndex + 1;                                // 701
        if (newFocusIndex >= items.length) {                           // 702
          newFocusIndex = 0;                                           // 703
        }                                                              //
      } else if (event.keyCode == 13) {                                //
        // Enter key                                                   //
        event.preventDefault();                                        // 706
        activateTargetTab(event, template);                            // 707
      }                                                                //
      if (newFocusIndex != null) {                                     // 709
        items.attr("tabindex", "-1");                                  // 710
        newFocus = $(items[newFocusIndex]);                            // 711
        newFocus.attr("tabindex", "0");                                // 712
        newFocus.focus();                                              // 713
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.shareableLinkTab.events({                                   // 718
    "change .share-token-role": function (event, instance) {           // 719
      var success = instance.completionState.get().success;            // 720
      if (success) {                                                   // 721
        var roleList = event.target;                                   // 722
        var assignment;                                                // 723
        if (roleList) {                                                // 724
          assignment = { roleId: roleList.selectedIndex };             // 725
        } else {                                                       //
          assignment = { none: null };                                 // 727
        }                                                              //
        Meteor.call("updateApiToken", success.id, { roleAssignment: assignment }, function (error) {
          if (error) {                                                 // 730
            console.error(error.stack);                                // 731
          }                                                            //
        });                                                            //
      }                                                                //
    },                                                                 //
    "change .label": function (event, instance) {                      // 736
      var success = instance.completionState.get().success;            // 737
      if (success) {                                                   // 738
        var label = event.target.value;                                // 739
        Meteor.call("updateApiToken", success.id, { petname: label }, function (error) {
          if (error) {                                                 // 741
            console.error(error.stack);                                // 742
          }                                                            //
        });                                                            //
      }                                                                //
    },                                                                 //
    "submit form.new-share-token": function (event, instance) {        // 747
      event.preventDefault();                                          // 748
      if (!instance.completionState.get().clear) {                     // 749
        return;                                                        // 750
      }                                                                //
      var currentGrain = getActiveGrain(globalGrains.get());           // 752
      var grainId = currentGrain.grainId();                            // 753
      var roleList = event.target.getElementsByClassName("share-token-role")[0];
      var assignment;                                                  // 755
      if (roleList) {                                                  // 756
        assignment = { roleId: roleList.selectedIndex };               // 757
      } else {                                                         //
        assignment = { none: null };                                   // 759
      }                                                                //
      instance.completionState.set({ "pending": true });               // 761
      Meteor.call("newApiToken", { identityId: currentGrain.identityId() }, grainId, event.target.getElementsByClassName("label")[0].value, assignment, { webkey: { forSharing: true } }, function (error, result) {
        if (error) {                                                   // 766
          console.error(error.stack);                                  // 767
        } else {                                                       //
          result.url = getOrigin() + "/shared/" + result.token;        // 769
          instance.completionState.set({ "success": result });         // 770
          // On the next render, .copy-me will exist, and we should focus it then.
          Meteor.defer(function () {                                   // 772
            var element = instance.find(".copy-me");                   // 773
            element.focus();                                           // 774
            selectElementContents(element);                            // 775
          });                                                          //
        }                                                              //
      });                                                              //
    },                                                                 //
    "click .reset-share-token": function (event, instance) {           // 780
      instance.completionState.set({ clear: true });                   // 781
      instance.find("form").reset();                                   // 782
      instance.find("form option[data-default-selected=true]").selected = true;
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainPowerboxRequestPopup.events({                          // 787
    "submit #powerbox-request-form": function (event) {                // 788
      event.preventDefault();                                          // 789
      var powerboxRequestInfo = this;                                  // 790
      var identityId = getActiveGrain(globalGrains.get()).identityId();
      Meteor.call("finishPowerboxRequest", event.target.token.value, powerboxRequestInfo.saveLabel, identityId, powerboxRequestInfo.grainId, function (err, token) {
        if (err) {                                                     // 795
          powerboxRequestInfo.error.set(err.toString());               // 796
        } else {                                                       //
          powerboxRequestInfo.source.postMessage({                     // 798
            rpcId: powerboxRequestInfo.rpcId,                          // 800
            token: token                                               // 801
          }, powerboxRequestInfo.origin);                              //
          powerboxRequestInfo.closer.close();                          // 803
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainPowerboxOfferPopup.events({                            // 810
    "click button.dismiss": function (event, instance) {               // 811
      var sessionId = instance.data.sessionId;                         // 812
      if (sessionId) {                                                 // 813
        Meteor.call("finishPowerboxOffer", sessionId, function (err) {
          // TODO(someday): display the error nicely to the user       //
          if (err) {                                                   // 816
            console.error(err);                                        // 817
          }                                                            //
        });                                                            //
      } else {                                                         //
        // TODO(cleanup): This path is used by the admin UI. This is really hacky, though.
        Iron.controller().state.set("powerboxOfferUrl", null);         // 822
      }                                                                //
    },                                                                 //
    "click .copy-me": selectTargetContents,                            // 825
    "focus .copy-me": selectTargetContents                             // 826
  });                                                                  //
                                                                       //
  Template.grainSharePopup.helpers({                                   // 829
    "currentGrain": function () {                                      // 830
      return getActiveGrain(globalGrains.get());                       // 831
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainView.helpers({                                         // 835
    unpackedGrainState: function () {                                  // 836
      return mapGrainStateToTemplateData(this);                        // 837
    },                                                                 //
    identityPickerData: function () {                                  // 839
      var identities = SandstormDb.getUserIdentityIds(Meteor.user()).map(function (id) {
        var identity = Meteor.users.findOne({ _id: id });              // 841
        if (identity) {                                                // 842
          SandstormDb.fillInProfileDefaults(identity);                 // 843
          SandstormDb.fillInIntrinsicName(identity);                   // 844
          SandstormDb.fillInPictureUrl(identity);                      // 845
          return identity;                                             // 846
        }                                                              //
      });                                                              //
      var grain = getActiveGrain(globalGrains.get());                  // 849
      return { identities: identities,                                 // 850
        onPicked: function (identityId) {                              // 851
          grain.revealIdentity(identityId);                            // 851
        } };                                                           //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grain.helpers({                                             // 855
    currentGrain: function () {                                        // 856
      return getActiveGrain(globalGrains.get());                       // 857
    },                                                                 //
    isOwner: function () {                                             // 859
      var current = getActiveGrain(globalGrains.get());                // 860
      return current && current.isOwner();                             // 861
    },                                                                 //
                                                                       //
    grainSize: function () {                                           // 864
      var current = getActiveGrain(globalGrains.get());                // 865
      var sizeBytes = current && current.size();                       // 866
      if (sizeBytes !== undefined) {                                   // 867
        return "(" + prettySize(sizeBytes) + ")";                      // 868
      } else {                                                         //
        return "";                                                     // 870
      }                                                                //
    },                                                                 //
                                                                       //
    setGrainWindowTitle: function () {                                 // 874
      var current = getActiveGrain(globalGrains.get());                // 875
      if (current) {                                                   // 876
        current.updateDocumentTitle();                                 // 877
      }                                                                //
    },                                                                 //
                                                                       //
    displayWebkeyButton: function () {                                 // 881
      // TODO: figure out what this should do                          //
      return Meteor.userId() || !this.oldSharingModel;                 // 883
    },                                                                 //
                                                                       //
    showPowerboxOffer: function () {                                   // 886
      var current = getActiveGrain(globalGrains.get());                // 887
      if (current) {                                                   // 888
        var session = Sessions.findOne({ _id: current.sessionId() }, { fields: { powerboxView: 1 } });
        return session && session.powerboxView && !!session.powerboxView.offer;
      }                                                                //
      return false;                                                    // 892
    }                                                                  //
                                                                       //
  });                                                                  //
                                                                       //
  Template.grainTitle.helpers({                                        // 897
    title: function () {                                               // 898
      var grain = getActiveGrain(globalGrains.get());                  // 899
      return grain && grain.title() || "Untitled grain";               // 900
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainApiTokenPopup.helpers({                                // 904
    displayToken: function () {                                        // 905
      return !this.revoked && !this.expiresIfUnused && !this.parentToken;
    },                                                                 //
    existingTokens: function () {                                      // 908
      var current = getActiveGrain(globalGrains.get());                // 909
      return current && ApiTokens.find({ grainId: current.grainId(),   // 910
        identityId: current.identityId(),                              // 911
        forSharing: { $ne: true },                                     // 912
        $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }],
        expiresIfUnused: null });                                      // 915
    },                                                                 //
    generatedApiToken: function () {                                   // 917
      var current = getActiveGrain(globalGrains.get());                // 918
      return current && current.generatedApiToken();                   // 919
    },                                                                 //
    generatedApiTokenPending: function () {                            // 921
      var current = getActiveGrain(globalGrains.get());                // 922
      return (current && current.generatedApiToken()) == "pending";    // 923
    },                                                                 //
    viewInfo: function () {                                            // 925
      var current = getActiveGrain(globalGrains.get());                // 926
      return current && current.viewInfo();                            // 927
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.whoHasAccessPopup.onCreated(function () {                   // 931
    var instance = this;                                               // 932
    var currentGrain = getActiveGrain(globalGrains.get());             // 933
    instance.identityId = currentGrain.identityId();                   // 934
    instance.grainId = currentGrain.grainId();                         // 935
    instance.transitiveShares = new ReactiveVar(null);                 // 936
    this.resetTransitiveShares = function () {                         // 937
      Meteor.call("transitiveShares", instance.identityId, instance.grainId, function (error, downstream) {
        if (error) {                                                   // 940
          console.error(error.stack);                                  // 941
        } else {                                                       //
          var sharesByRecipient = {};                                  // 943
          downstream.forEach(function (token) {                        // 944
            if (Match.test(token.owner, { user: Match.ObjectIncluding({ identityId: String }) })) {
              var recipient = token.owner.user.identityId;             // 946
              if (!sharesByRecipient[recipient]) {                     // 947
                sharesByRecipient[recipient] = { recipient: recipient, shares: [] };
              }                                                        //
              var shares = sharesByRecipient[recipient].shares;        // 950
              if (!shares.some(function (share) {                      // 951
                return share.identityId === token.identityId;          // 951
              })) {                                                    //
                sharesByRecipient[recipient].shares.push(token);       // 952
              }                                                        //
            }                                                          //
          });                                                          //
          var result = _.values(sharesByRecipient);                    // 956
          if (result.length == 0) {                                    // 957
            result = { empty: true };                                  // 958
          }                                                            //
          instance.transitiveShares.set(result);                       // 960
        }                                                              //
      });                                                              //
    };                                                                 //
    this.resetTransitiveShares();                                      // 964
  });                                                                  //
                                                                       //
  Template.whoHasAccessPopup.events({                                  // 967
    "change .share-token-role": function (event, instance) {           // 968
      var roleList = event.target;                                     // 969
      var assignment;                                                  // 970
      if (roleList) {                                                  // 971
        assignment = { roleId: roleList.selectedIndex };               // 972
      } else {                                                         //
        assignment = { none: null };                                   // 974
      }                                                                //
      Meteor.call("updateApiToken", roleList.getAttribute("data-token-id"), { roleAssignment: assignment }, function (error) {
        if (error) {                                                   // 978
          console.error(error.stack);                                  // 979
        }                                                              //
      });                                                              //
    },                                                                 //
    "click button.revoke-token": function (event, instance) {          // 983
      Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
      instance.resetTransitiveShares();                                // 986
    },                                                                 //
    "click .token-petname": function (event) {                         // 988
      // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
      //   to edit the petname in place.                               //
      var petname = window.prompt("Set new label:", this.petname);     // 991
      if (petname) {                                                   // 992
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  function isEmptyPermissionSet(permissionSet) {                       // 999
    if (!permissionSet) {                                              // 1000
      return true;                                                     // 1001
    }                                                                  //
    for (var ii = 0; ii < permissionSet.length; ++ii) {                // 1003
      if (permissionSet[ii]) {                                         // 1004
        return false;                                                  // 1005
      }                                                                //
    }                                                                  //
    return true;                                                       // 1008
  }                                                                    //
                                                                       //
  Template.whoHasAccessPopup.helpers({                                 // 1011
    existingShareTokens: function () {                                 // 1012
      if (Meteor.userId()) {                                           // 1013
        return ApiTokens.find({ grainId: Template.instance().grainId,  // 1014
          identityId: Template.instance().identityId,                  // 1015
          forSharing: true,                                            // 1016
          $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }]
        }).fetch();                                                    //
      }                                                                //
    },                                                                 //
    getPetname: function () {                                          // 1022
      if (this.petname) {                                              // 1023
        return this.petname;                                           // 1024
      } else {                                                         //
        return "Unlabeled Link";                                       // 1026
      }                                                                //
    },                                                                 //
    displayName: function (identityId) {                               // 1029
      var name = DisplayNames.findOne(identityId);                     // 1030
      if (name) {                                                      // 1031
        return name.displayName;                                       // 1032
      } else {                                                         //
        return "Unknown User (" + identityId.slice(0, 16) + ")";       // 1034
      }                                                                //
    },                                                                 //
    transitiveShares: function () {                                    // 1037
      return Template.instance().transitiveShares.get();               // 1038
    },                                                                 //
    indexedRoles: function () {                                        // 1040
      var result = [];                                                 // 1041
      var instance = Template.instance();                              // 1042
      var currentGrain = getActiveGrain(globalGrains.get());           // 1043
      var roles = currentGrain.viewInfo().roles;                       // 1044
      for (var ii = 0; ii < roles.length; ++ii) {                      // 1045
        result.push({ idx: ii, title: roles[ii].title, verbPhrase: roles[ii].verbPhrase });
      }                                                                //
      return result;                                                   // 1048
    },                                                                 //
    roleText: function () {                                            // 1050
      if (this.verbPhrase) {                                           // 1051
        return this.verbPhrase.defaultText;                            // 1052
      } else {                                                         //
        return "is " + this.title.defaultText;                         // 1054
      }                                                                //
    },                                                                 //
    hasCustomRole: function (token) {                                  // 1057
      var role = token.roleAssignment;                                 // 1058
      if ("roleId" in role && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
        return false;                                                  // 1062
      }                                                                //
      return true;                                                     // 1064
    },                                                                 //
    hasCurrentRole: function (token) {                                 // 1066
      var role = token.roleAssignment;                                 // 1067
      if ("roleId" in role && role.roleId == this.idx && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
        return true;                                                   // 1071
      }                                                                //
      return false;                                                    // 1073
    },                                                                 //
    displayToken: function () {                                        // 1075
      return !this.revoked && !this.expiresIfUnused && !this.parentToken;
    },                                                                 //
    viewInfo: function () {                                            // 1078
      var activeGrain = getActiveGrain(globalGrains.get());            // 1079
      return activeGrain && activeGrain.viewInfo();                    // 1080
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.selectRole.helpers({                                        // 1084
    roleText: function () {                                            // 1085
      if (this.verbPhrase) {                                           // 1086
        return this.verbPhrase.defaultText;                            // 1087
      } else {                                                         //
        return "is " + this.title.defaultText;                         // 1089
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.shareableLinkTab.onCreated(function () {                    // 1094
    this.completionState = new ReactiveVar({ clear: true });           // 1095
  });                                                                  //
                                                                       //
  Template.emailInviteTab.onCreated(function () {                      // 1098
    this.completionState = new ReactiveVar({ clear: true });           // 1099
    this.contacts = new ReactiveVar([]);                               // 1100
  });                                                                  //
                                                                       //
  Template.shareableLinkTab.helpers({                                  // 1103
    completionState: function () {                                     // 1104
      var instance = Template.instance();                              // 1105
      return instance.completionState.get();                           // 1106
    }                                                                  //
  });                                                                  //
  Template.emailInviteTab.helpers({                                    // 1109
    completionState: function () {                                     // 1110
      var instance = Template.instance();                              // 1111
      return instance.completionState.get();                           // 1112
    },                                                                 //
    contacts: function () {                                            // 1114
      return Template.instance().contacts;                             // 1115
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.emailInviteTab.events({                                     // 1119
    "submit form.email-invite": function (event, instance) {           // 1120
      event.preventDefault();                                          // 1121
      return false;                                                    // 1122
    },                                                                 //
    "click form.email-invite button": function (event, instance) {     // 1124
      event.preventDefault();                                          // 1125
      if (!instance.completionState.get().clear) {                     // 1126
        return;                                                        // 1127
      }                                                                //
      var grainId = instance.data.grainId;                             // 1129
      var title = instance.data.title;                                 // 1130
                                                                       //
      var roleList = instance.find(".share-token-role");               // 1132
      var assignment;                                                  // 1133
      if (roleList) {                                                  // 1134
        assignment = { roleId: roleList.selectedIndex };               // 1135
      } else {                                                         //
        assignment = { none: null };                                   // 1137
      }                                                                //
      var message = instance.find(".personal-message").value;          // 1139
      instance.completionState.set({ pending: true });                 // 1140
                                                                       //
      var currentGrain = getActiveGrain(globalGrains.get());           // 1142
                                                                       //
      // HTML-escape the message.                                      //
      var div = document.createElement('div');                         // 1145
      div.appendChild(document.createTextNode(message));               // 1146
      var htmlMessage = div.innerHTML.replace(/\n/g, "<br>");          // 1147
                                                                       //
      var contacts = instance.contacts.get();                          // 1149
      var emails = instance.find("input.emails");                      // 1150
      var emailsValue = emails.value;                                  // 1151
      if (emailsValue) {                                               // 1152
        if (emailsValue.match(/.+\@.+\..+/)) {                         // 1153
          contacts.push({                                              // 1154
            _id: emailsValue,                                          // 1155
            profile: {                                                 // 1156
              service: "email",                                        // 1157
              name: emailsValue,                                       // 1158
              intrinsicName: emailsValue,                              // 1159
              pictureUrl: "/email.svg"                                 // 1160
            },                                                         //
            isDefault: true                                            // 1162
          });                                                          //
          instance.contacts.set(contacts);                             // 1164
          emails.value = "";                                           // 1165
        } else {                                                       //
          instance.completionState.set({ error: "Invalid text in contact input box: " + emailsValue });
          return;                                                      // 1169
        }                                                              //
      }                                                                //
                                                                       //
      Meteor.call("inviteUsersToGrain", getOrigin(), currentGrain.identityId(), grainId, title, assignment, contacts, { text: message, html: htmlMessage }, function (error, result) {
        if (error) {                                                   // 1176
          instance.completionState.set({ error: error.toString() });   // 1177
        } else {                                                       //
          if (result.failures.length > 0) {                            // 1179
            var message = "Failed to send to: ";                       // 1180
            for (var ii = 0; ii < result.failures.length; ++ii) {      // 1181
              console.error(result.failures[ii].error);                // 1182
              if (ii != 0) {                                           // 1183
                message += ", ";                                       // 1184
              }                                                        //
              message += result.failures[ii].contact.profile.name;     // 1186
              var warning = result.failures[ii].warning;               // 1187
              if (warning) {                                           // 1188
                message += ". " + warning;                             // 1189
              }                                                        //
            }                                                          //
            instance.completionState.set({ error: message });          // 1192
          } else {                                                     //
            instance.completionState.set({ success: "success" });      // 1194
          }                                                            //
        }                                                              //
      });                                                              //
    },                                                                 //
    "click .reset-invite": function (event, instance) {                // 1199
      instance.contacts.set([]);                                       // 1200
      instance.completionState.set({ clear: true });                   // 1201
      instance.find("form").reset();                                   // 1202
      instance.find("form option[data-default-selected=true]").selected = true;
    },                                                                 //
    "click .start-over-invite": function (event, instance) {           // 1205
      instance.completionState.set({ clear: true });                   // 1206
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainPowerboxOfferPopup.helpers({                           // 1210
    powerboxOfferUrl: function () {                                    // 1211
      if (this.powerboxOfferUrl) {                                     // 1212
        // TODO(cleanup): This path is used by the admin UI. This is really hacky, though.
        return this.powerboxOfferUrl;                                  // 1214
      }                                                                //
                                                                       //
      var activeGrain = getActiveGrain(globalGrains.get());            // 1217
      var session = Sessions.findOne({ _id: activeGrain.sessionId() }, { fields: { powerboxView: 1 } });
      return session && session.powerboxView && session.powerboxView.offer;
    }                                                                  //
  });                                                                  //
                                                                       //
  // Send a keep-alive for each grain every now and then.              //
  Meteor.setInterval(function () {                                     // 1224
    var grains = globalGrains.get();                                   // 1225
    if (!grains) return;                                               // 1226
                                                                       //
    // Meteor has an exponential backoff of up to 5 minutes for reconnect. This is unnacceptable
    // for us, since we rely on Sessions being re-established in under 60s.
    if (Meteor.status().status === "waiting") {                        // 1230
      console.log("Sandstorm is trying to reconnect...");              // 1231
      Meteor.reconnect();                                              // 1232
    }                                                                  //
                                                                       //
    grains.forEach(function (grain) {                                  // 1235
      if (grain.sessionId()) {                                         // 1236
        // TODO(soon):  Investigate what happens in background tabs.  Maybe arrange to re-open the
        //   app if it dies while in the background.                   //
        Meteor.call("keepSessionAlive", grain.sessionId(), function (error, result) {
          // Sessions will automatically resume if possible, otherwise they will refresh.
        });                                                            //
      }                                                                //
    });                                                                //
  }, 60000);                                                           //
                                                                       //
  // Message handler for Sandstorm's client-side postMessage API.      //
  Meteor.startup(function () {                                         // 1247
    var messageListener = function (event) {                           // 1248
      if (event.origin === getOrigin()) {                              // 1249
        // Meteor likes to postMessage() to itself sometimes, so we ignore these messages.
        return;                                                        // 1251
      }                                                                //
                                                                       //
      // Look up the grain index of which grain this postmessage came from, so we can map behavior
      // into that particular grain's state                            //
      var grains = globalGrains.get();                                 // 1256
      var senderGrainIndex = grainOriginToIndex(grains, event.origin);
      if (senderGrainIndex == -1) {                                    // 1258
        // We got a postMessage from an origin that is not a grain we currently believe is open.
        // Ignore it. (It may be aimed at some other message listener registered elsewhere...)
        return;                                                        // 1261
      }                                                                //
      var senderGrain = grains[senderGrainIndex];                      // 1263
                                                                       //
      if (event.data.setPath || event.data.setPath === '') {           // 1265
        var path = event.data.setPath || '/';                          // 1266
        check(path, String);                                           // 1267
        check(path.charAt(0), '/');                                    // 1268
        // TODO(security): More sanitization of this path. E.g. reject "/../../".
        senderGrain.setPath(path);                                     // 1270
      } else if (event.data.startSharing) {                            //
        // Allow the current grain to request that the "Share Access" menu be shown.
        // Only show this popup if no other popup is currently active.
        // TODO(security): defend against malicious apps spamming this call, blocking all other UI.
        var currentGrain = getActiveGrain(globalGrains.get());         // 1275
        if (senderGrain === currentGrain && !globalTopbar.isPopupOpen()) {
          globalTopbar.openPopup("share");                             // 1277
        }                                                              //
      } else if (event.data.setTitle || event.data.setTitle === '') {  //
        senderGrain.setFrameTitle(event.data.setTitle);                // 1280
      } else if (event.data.renderTemplate) {                          //
        // Request creation of a single-use template with a privileged API token.
        // Why?  Apps should not be able to obtain capabilities-as-keys to
        // themselves directly, because those can be leaked through an
        // arbitrary bit stream or covert channel.  However, apps often need a
        // way to provide instructions to users to copy/paste with some
        // privileged token contained within.  By providing this templating in
        // the platform, we can ensure that the token is only visible to the
        // shell's origin.                                             //
        var call = event.data.renderTemplate;                          // 1290
        check(call, Object);                                           // 1291
        var rpcId = call.rpcId;                                        // 1292
        try {                                                          // 1293
          check(call, { rpcId: String, template: String, petname: Match.Optional(String),
            roleAssignment: Match.Optional(roleAssignmentPattern),     // 1295
            forSharing: Match.Optional(Boolean) });                    // 1296
        } catch (error) {                                              //
          event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
          return;                                                      // 1299
        }                                                              //
        var template = call.template;                                  // 1301
        var petname = "connected external app";                        // 1302
        if (call.petname) {                                            // 1303
          petname = call.petname;                                      // 1304
        }                                                              //
        var assignment = { allAccess: null };                          // 1306
        if (call.roleAssignment) {                                     // 1307
          assignment = call.roleAssignment;                            // 1308
        }                                                              //
        var forSharing = call.forSharing ? call.forSharing : false;    // 1310
        // Tokens expire by default in 5 minutes from generation date  //
        var selfDestructDuration = 5 * 60 * 1000;                      // 1312
                                                                       //
        var provider;                                                  // 1314
        if (Router.current().route.getName() === "shared") {           // 1315
          provider = { rawParentToken: Router.current().params.token };
        } else {                                                       //
          provider = { identityId: senderGrain.identityId() };         // 1318
        }                                                              //
        var owner = { webkey: { forSharing: forSharing,                // 1320
            expiresIfUnusedDuration: selfDestructDuration } };         // 1321
                                                                       //
        Meteor.call("newApiToken", provider, senderGrain.grainId(), petname, assignment, owner, function (error, result) {
          if (error) {                                                 // 1325
            event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
          } else {                                                     //
            var tokenId = result.token;                                // 1328
            // Generate random key id2.                                //
            var id2 = Random.secret();                                 // 1330
            // Store apitoken id1 and template in session storage in the offer
            // template namespace under key id2.                       //
            var key = "offerTemplate" + id2;                           // 1333
            var renderedTemplate = template.replace(/\$API_TOKEN/g, tokenId).replace(/\$API_HOST/g, makeWildcardHost("api"));
            sessionStorage.setItem(key, JSON.stringify({               // 1336
              "token": tokenId,                                        // 1337
              "renderedTemplate": renderedTemplate,                    // 1338
              "expires": Date.now() + selfDestructDuration             // 1339
            }));                                                       //
            sessionStorage.setItem("apiHost", makeWildcardHost("api"));
                                                                       //
            // Send message to event.source with URL containing id2    //
            templateLink = window.location.origin + "/offer-template.html#" + id2;
            event.source.postMessage({ rpcId: rpcId, uri: templateLink }, event.origin);
          }                                                            //
        });                                                            //
      } else if (event.data.powerboxRequest) {                         //
        // TODO(now): make this work with GrainView                    //
        var powerboxRequest = event.data.powerboxRequest;              // 1351
        check(powerboxRequest, Object);                                // 1352
        var rpcId = powerboxRequest.rpcId;                             // 1353
                                                                       //
        var powerboxRequestInfo = {                                    // 1355
          source: event.source,                                        // 1356
          rpcId: rpcId,                                                // 1357
          grainId: senderGrain.grainId(),                              // 1358
          origin: event.origin,                                        // 1359
          saveLabel: powerboxRequest.saveLabel,                        // 1360
          error: new ReactiveVar(null)                                 // 1361
        };                                                             //
                                                                       //
        powerboxRequestInfo.closer = globalTopbar.addItem({            // 1364
          name: "request",                                             // 1365
          template: Template.grainPowerboxRequest,                     // 1366
          popupTemplate: Template.grainPowerboxRequestPopup,           // 1367
          data: new ReactiveVar(powerboxRequestInfo),                  // 1368
          startOpen: true,                                             // 1369
          onDismiss: function () {                                     // 1370
            powerboxRequestInfo.source.postMessage({                   // 1371
              rpcId: powerboxRequestInfo.rpcId,                        // 1373
              error: "User canceled request"                           // 1374
            }, powerboxRequestInfo.origin);                            //
            return "remove";                                           // 1376
          }                                                            //
        });                                                            //
      } else {                                                         //
        console.log("postMessage from app not understood: " + event.data);
        console.log(event);                                            // 1381
      }                                                                //
    };                                                                 //
                                                                       //
    window.addEventListener("message", messageListener, false);        // 1385
  });                                                                  //
}                                                                      //
                                                                       //
if (Meteor.isClient) {                                                 // 1389
  function maybeScrollLog() {                                          // 1390
    var elem = document.getElementById("grainLog");                    // 1391
    if (elem) {                                                        // 1392
      // The log already exists. It's about to be updated. Check if it's scrolled to the bottom
      // before the update.                                            //
      if (elem.scrollHeight - elem.scrollTop === elem.clientHeight) {  // 1395
        // Indeed, so we want to scroll it back to the bottom after the update.
        Tracker.afterFlush(function () {                               // 1397
          scrollLogToBottom(elem);                                     // 1397
        });                                                            //
      }                                                                //
    } else {                                                           //
      // No element exists yet, but it's probably about to be created, in which case we definitely
      // want to scroll it.                                            //
      Tracker.afterFlush(function () {                                 // 1402
        var elem2 = document.getElementById("grainLog");               // 1403
        if (elem2) scrollLogToBottom(elem2);                           // 1404
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  function scrollLogToBottom(elem) {                                   // 1409
    elem.scrollTop = elem.scrollHeight;                                // 1410
  }                                                                    //
}                                                                      //
                                                                       //
function makeGrainIdActive(grainId) {                                  // 1414
  var grains = globalGrains.get();                                     // 1415
  for (var i = 0; i < grains.length; i++) {                            // 1416
    var grain = grains[i];                                             // 1417
    if (grain.grainId() === grainId) {                                 // 1418
      if (!grain.isActive()) {                                         // 1419
        grain.setActive(true);                                         // 1420
      }                                                                //
    } else {                                                           //
      if (grain.isActive()) {                                          // 1423
        grain.setActive(false);                                        // 1424
      }                                                                //
    }                                                                  //
  }                                                                    //
}                                                                      //
                                                                       //
function getActiveGrain(grains) {                                      // 1430
  var idx = activeGrainIndex(grains);                                  // 1431
  return idx == -1 ? undefined : grains[idx];                          // 1432
}                                                                      //
                                                                       //
function activeGrainIndex(grains) {                                    // 1435
  for (var i = 0; i < grains.length; i++) {                            // 1436
    if (grains[i].isActive()) {                                        // 1437
      return i;                                                        // 1438
    }                                                                  //
  }                                                                    //
  return -1;                                                           // 1441
}                                                                      //
                                                                       //
function grainOriginToIndex(grains, origin) {                          // 1444
  for (var i = 0; i < grains.length; i++) {                            // 1445
    if (grains[i].origin() === origin) {                               // 1446
      return i;                                                        // 1447
    }                                                                  //
  }                                                                    //
  return -1;                                                           // 1450
}                                                                      //
                                                                       //
function grainIdToIndex(grains, grainId) {                             // 1453
  for (var i = 0; i < grains.length; i++) {                            // 1454
    var grain = grains[i];                                             // 1455
    if (grains[i].grainId() === grainId) {                             // 1456
      return i;                                                        // 1457
    }                                                                  //
  }                                                                    //
  return -1;                                                           // 1460
}                                                                      //
                                                                       //
function mapGrainStateToTemplateData(grainState) {                     // 1463
  var templateData = {                                                 // 1464
    grainId: grainState.grainId(),                                     // 1465
    active: grainState.isActive(),                                     // 1466
    title: grainState.title(),                                         // 1467
    error: grainState.error(),                                         // 1468
    appOrigin: grainState.origin(),                                    // 1469
    hasNotLoaded: !grainState.hasLoaded(),                             // 1470
    sessionId: grainState.sessionId(),                                 // 1471
    originalPath: grainState._originalPath,                            // 1472
    interstitial: grainState.shouldShowInterstitial(),                 // 1473
    token: grainState.token(),                                         // 1474
    viewInfo: grainState.viewInfo()                                    // 1475
  };                                                                   //
  return templateData;                                                 // 1477
}                                                                      //
                                                                       //
GrainLog = new Mongo.Collection("grainLog");                           // 1480
// Pseudo-collection created by subscribing to "grainLog", implemented in proxy.js.
                                                                       //
Router.map(function () {                                               // 1483
  this.route("apps", {                                                 // 1484
    path: "/apps",                                                     // 1485
    waitOn: function () {                                              // 1486
      return globalSubs;                                               // 1486
    },                                                                 //
    data: function () {                                                // 1487
      if (!this.ready()) return;                                       // 1488
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1489
        Router.go("root", {}, { replaceState: true });                 // 1490
      }                                                                //
                                                                       //
      return new SandstormAppList(globalDb, globalQuotaEnforcer);      // 1493
    }                                                                  //
  });                                                                  //
  this.route("newGrainRedirect", {                                     // 1496
    // The path /grain/new used to be where you would go to create new grains.
    // Its functionality has been superceded by the apps route, so redirect in
    // case anyone has the old link saved somewhere.                   //
    path: "/grain/new",                                                // 1500
    onBeforeAction: function () {                                      // 1501
      Router.go("apps", {}, { replaceState: true });                   // 1502
    }                                                                  //
  });                                                                  //
  this.route("grains", {                                               // 1505
    path: "/grain",                                                    // 1506
    waitOn: function () {                                              // 1507
      return globalSubs;                                               // 1507
    },                                                                 //
    data: function () {                                                // 1508
      if (!this.ready()) return;                                       // 1509
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1510
        Router.go("root", {}, { replaceState: true });                 // 1511
      }                                                                //
                                                                       //
      return new SandstormGrainListPage(globalDb, globalQuotaEnforcer);
    }                                                                  //
  });                                                                  //
  this.route("grain", {                                                // 1517
    path: "/grain/:grainId/:path(.*)?",                                // 1518
    loadingTemplate: "loadingNoMessage",                               // 1519
                                                                       //
    waitOn: function () {                                              // 1521
      return globalSubs;                                               // 1522
    },                                                                 //
                                                                       //
    onBeforeAction: function () {                                      // 1525
      // Run this hook only once.                                      //
      if (this.state.get("beforeActionHookRan")) {                     // 1527
        return this.next();                                            // 1527
      }                                                                //
      this.state.set("beforeActionHookRan", true);                     // 1528
      var grainId = this.params.grainId;                               // 1529
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var grains = globalGrains.get();                                 // 1531
      var grainIndex = grainIdToIndex(grains, grainId);                // 1532
      if (grainIndex == -1) {                                          // 1533
        // The element we need to attach our Blaze view to may not exist yet.
        // In that case, defer creating the GrainView until we're sure it's
        // had a chance to render.                                     //
        var openView = (function () {                                  // 1537
          function openView() {                                        // 1537
            var mainContentElement = document.querySelector("body>.main-content");
            if (mainContentElement) {                                  // 1539
              var grains = globalGrains.get();                         // 1540
              var grainToOpen = new GrainView(grainId, path, undefined, mainContentElement);
              grainToOpen.openSession();                               // 1542
              grainIndex = grains.push(grainToOpen) - 1;               // 1543
              globalGrains.set(grains);                                // 1544
              makeGrainIdActive(grainId);                              // 1545
            } else {                                                   //
              Meteor.defer(openView);                                  // 1547
            }                                                          //
          }                                                            //
                                                                       //
          return openView;                                             //
        })();                                                          //
        openView();                                                    // 1550
      } else {                                                         //
        makeGrainIdActive(grainId);                                    // 1552
      }                                                                //
      this.next();                                                     // 1554
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 1557
      this.state.set("beforeActionHookRan", undefined);                // 1558
      globalGrains.get().forEach(function (grain) {                    // 1559
        if (grain.isActive()) {                                        // 1560
          grain.setActive(false);                                      // 1561
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("/shared/:token/:path(.*)?", {                            // 1567
    name: "shared",                                                    // 1568
    template: "grain",                                                 // 1569
    loadingTemplate: "loadingNoMessage",                               // 1570
                                                                       //
    waitOn: function () {                                              // 1572
      return [Meteor.subscribe("devPackages"), Meteor.subscribe("tokenInfo", this.params.token), Meteor.subscribe("grainsMenu")
      // This subscription gives us the data we need for deciding whether to automatically reveal
      // our identity.                                                 //
      // TODO(soon): Subscribe to contacts instead.                    //
      ];                                                               //
    },                                                                 //
                                                                       //
    onBeforeAction: function () {                                      // 1584
      // Run this hook only once. We could accomplish the same thing by using the `onRun()` hook
      // and waiting for `this.ready()`, but for some reason that fails in the case when a user
      // logs in while visiting a /shared/ link.                       //
      if (this.state.get("beforeActionHookRan")) {                     // 1588
        return this.next();                                            // 1588
      }                                                                //
      this.state.set("beforeActionHookRan", true);                     // 1589
                                                                       //
      var token = this.params.token;                                   // 1591
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var hash = this.params.hash;                                     // 1593
                                                                       //
      var tokenInfo = TokenInfo.findOne({ _id: token });               // 1595
      if (tokenInfo && tokenInfo.invalidToken) {                       // 1596
        this.state.set("invalidToken", true);                          // 1597
      } else {                                                         //
        var grainId = tokenInfo.grainId;                               // 1599
        var grains = globalGrains.get();                               // 1600
        var grainIndex = grainIdToIndex(grains, grainId);              // 1601
        if (grainIndex == -1) {                                        // 1602
          var openView = (function () {                                // 1603
            function openView() {                                      // 1603
              var mainContentElement = document.querySelector("body>.main-content");
              if (mainContentElement) {                                // 1605
                var grains = globalGrains.get();                       // 1606
                var grainToOpen = new GrainView(grainId, path, tokenInfo, mainContentElement);
                grainToOpen.openSession();                             // 1608
                grainIndex = grains.push(grainToOpen) - 1;             // 1609
                globalGrains.set(grains);                              // 1610
                makeGrainIdActive(grainId);                            // 1611
              } else {                                                 //
                Meteor.defer(openView);                                // 1613
              }                                                        //
            }                                                          //
                                                                       //
            return openView;                                           //
          })();                                                        //
          openView();                                                  // 1616
        } else {                                                       //
          makeGrainIdActive(grainId);                                  // 1618
        }                                                              //
      }                                                                //
      this.next();                                                     // 1621
    },                                                                 //
                                                                       //
    action: function () {                                              // 1624
      if (this.state.get("invalidToken")) {                            // 1625
        this.render("invalidToken", { data: { token: this.params.token } });
      } else {                                                         //
        this.render();                                                 // 1628
      }                                                                //
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 1632
      this.state.set("beforeActionHookRan", undefined);                // 1633
      this.state.set("invalidToken", undefined);                       // 1634
      globalGrains.get().forEach(function (grain) {                    // 1635
        if (grain.isActive()) {                                        // 1636
          grain.setActive(false);                                      // 1637
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("grainLog", {                                             // 1643
    path: "/grainlog/:grainId",                                        // 1644
    layoutTemplate: "noLayout",                                        // 1645
                                                                       //
    waitOn: function () {                                              // 1647
      return [Meteor.subscribe("grainTopBar", this.params.grainId), Meteor.subscribe("grainLog", this.params.grainId)];
    },                                                                 //
                                                                       //
    data: function () {                                                // 1654
      if (this.ready()) {                                              // 1655
        maybeScrollLog();                                              // 1656
        var grain = Grains.findOne(this.params.grainId);               // 1657
        return {                                                       // 1658
          title: grain ? grain.title : "(deleted grain)",              // 1659
          html: AnsiUp.ansi_to_html(GrainLog.find({}, { $sort: { _id: 1 } }).map(function (entry) {
            return entry.text;                                         // 1661
          }).join(""), { use_classes: true })                          //
        };                                                             //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("appDetails", {                                           // 1668
    path: "/apps/:appId",                                              // 1669
    template: "appDetails",                                            // 1670
    waitOn: function () {                                              // 1671
      return globalSubs.concat(Meteor.subscribe("appIndex", this.params.appId));
    },                                                                 //
    data: function () {                                                // 1675
      return new SandstormAppDetails(globalDb, globalQuotaEnforcer, this.params.appId);
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=grain.js.map
