(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, isDemoUser, isSignedUp, isSignedUpOrDemo, successfulReferralsCount, getUserQuota, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, sendReferralProgramNotification, roleAssignmentPattern, SandstormDb, i, appName, addStaticAsset, verifiedEmails, unverifiedEmails, userPictureUrl, fetchPicture;

(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/db.js                                                                               //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
// This file defines the database schema.                                                                    // 17
                                                                                                             // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every               // 19
// query it makes, so you can see if it's doing queries too often, etc.                                      // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                      // 21
  var oldFind = Mongo.Collection.prototype.find;                                                             // 22
  Mongo.Collection.prototype.find = function () {                                                            // 23
    console.log(this._prefix, arguments);                                                                    // 24
    return oldFind.apply(this, arguments);                                                                   // 25
  }                                                                                                          // 26
}                                                                                                            // 27
                                                                                                             // 28
// Users = new Mongo.Collection("users");                                                                    // 29
// The users collection is special and can be accessed through `Meteor.users`.                               // 30
// See https://docs.meteor.com/#/full/meteor_users.                                                          // 31
//                                                                                                           // 32
// There are two distinct types of entries in the users collection: identities and accounts. An              // 33
// identity contains personal profile information and typically includes some intrinsic method for           // 34
// authenticating as the owner of that information.                                                          // 35
//                                                                                                           // 36
// An account is an owner of app actions, grains, contacts, notifications, and payment info.                 // 37
// Each account can have multiple identities linked to it. To log in as an account you must first            // 38
// authenticate as one of its linked identities.                                                             // 39
//                                                                                                           // 40
// Every user contains the following fields:                                                                 // 41
//   _id: Unique string ID. For accounts, this is random. For identities, this is the globally               // 42
//        stable SHA-256 ID of this identity, hex-encoded.                                                   // 43
//   createdAt: Date when this entry was added to the collection.                                            // 44
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                      // 45
//   services: Object containing login data used by Meteor authentication services.                          // 46
//   expires: Date when this user should be deleted. Only present for demo users.                            // 47
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.              // 48
//                     TODO(cleanup): Unlike other dates in our database, this is stored as a number         // 49
//                     rather than as a Date object. We should fix that.                                     // 50
//   appDemoId: If present and non-null, then the user is a demo user who arrived via an /appdemo/           // 51
//              link. This field contains the app ID of the app that the user started out demoing.           // 52
//              Unlike the `expires` field, this field is not cleared when the user upgrades from            // 53
//              being a demo user.                                                                           // 54
//                                                                                                           // 55
// Identity users additionally contain the following fields:                                                 // 56
//   profile: Object containing the data that will be shared with users and grains that come into            // 57
//            contact with this identity. Includes the following fields:                                     // 58
//       service: String containing the name of this identity's authentication method.                       // 59
//       name: String containing the chosen display name of the identity.                                    // 60
//       handle: String containing the identity's preferred handle.                                          // 61
//       picture: _id into the StaticAssets table for the identity's picture. If not present,                // 62
//                an identicon will be used.                                                                 // 63
//       pronoun: One of "male", "female", "neutral", or "robot".                                            // 64
//   unverifiedEmail: If present, a string containing an email address specified by the user.                // 65
//   referredBy: ID of the Account that referred this Identity.                                              // 66
//                                                                                                           // 67
// Account users additionally contain the following fields:                                                  // 68
//   loginIdentities: Array of identity objects, each of which may include the following fields.             // 69
//       id: The globally-stable SHA-256 ID of this identity, hex-encoded.                                   // 70
//   nonloginIdentities: Array of identity objects, of the same form as `loginIdentities`. We use            // 71
//                       a separate array here so that we can use a Mongo index to enforce the               // 72
//                       invariant that an identity only be a login identity for a single account.           // 73
//   primaryEmail: String containing this account's primary email address. Must be a verified adress         // 74
//                 of one of this account's linked identities. Call SandstormDb.getUserEmails()              // 75
//                 to do this checking automatically.                                                        // 76
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.        // 77
//   signupKey: If this is an invited user, then this field contains their signup key.                       // 78
//   signupNote: If the user was invited through a link, then this field contains the note that the          // 79
//               inviter admin attached to the key.                                                          // 80
//   signupEmail: If the user was invited by email, then this field contains the email address that          // 81
//                the invite was sent to.                                                                    // 82
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's          // 83
//                       terms of service.                                                                   // 84
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                             // 85
//   storageUsage: Number of bytes this user is currently storing.                                           // 86
//   payments: Object defined by payments module, if loaded.                                                 // 87
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                       // 88
//   referredByComplete: ID of the Account that referred this Account. If this is set, we                    // 89
//                        stop writing new referredBy values onto Identities for this account.               // 90
//   referredCompleteDate: The Date at which the completed referral occurred.                                // 91
//   referredIdentityIds: List of Identity IDs that this Account has referred. This is used for              // 92
//                        reliably determining which Identity's names are safe to display.                   // 93
//   experiments: Object where each field is an experiment that the user is in, and each value               // 94
//           is the parameters for that experiment. Typically, the value simply names which                  // 95
//           experiment group which the user is in, where "control" is one group. If an experiment           // 96
//           is not listed, then the user should not be considered at all for the purpose of that            // 97
//           experiment. Each experiment may define a point in time where users not already in the           // 98
//           experiment may be added to it and assigned to a group (for example, at user creation            // 99
//           time). Current experiments:                                                                     // 100
//       firstTimeBillingPrompt: Value is "control" or "test". Users are assigned to groups at               // 101
//               account creation on servers where billing is enabled (i.e. Oasis). Users in the             // 102
//               test group will see a plan selection dialog and asked to make an explitic choice            // 103
//               (possibly "free") before they can create grains (but not when opening someone               // 104
//               else's shared grain). The goal of the experiment is to determine whether this               // 105
//               prompt scares users away -- and also whether it increases paid signups.                     // 106
//   stashedOldUser: A complete copy of this user from before the accounts/identities migration.             // 107
//                   TODO(cleanup): Delete this field once we're sure it's safe to do so.                    // 108
                                                                                                             // 109
Packages = new Mongo.Collection("packages");                                                                 // 110
// Packages which are installed or downloading.                                                              // 111
//                                                                                                           // 112
// Each contains:                                                                                            // 113
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                          // 114
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"         // 115
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                      // 116
//       >1 = download byte count.                                                                           // 117
//   error:  If status is "failed", error message string.                                                    // 118
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                // 119
//   appId:  If status is "ready", the application ID string.  Packages representing different               // 120
//       versions of the same app have the same appId.  The spk tool defines the app ID format               // 121
//       and can cryptographically verify that a package belongs to a particular app ID.                     // 122
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package              // 123
//       collector should at some point check whether there are any other references and, if not,            // 124
//       delete the package.                                                                                 // 125
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.               // 126
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up          // 127
//     even if it has no users.                                                                              // 128
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                 // 129
//     packager.                                                                                             // 130
                                                                                                             // 131
DevPackages = new Mongo.Collection("devpackages");                                                           // 132
// List of packages currently made available via the dev tools running on the local machine.                 // 133
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool           // 134
// on the local machine to publish an under-development app to this server. That should only ever            // 135
// happen on developers' desktop machines.                                                                   // 136
//                                                                                                           // 137
// While a dev package is published, it automatically appears as installed by every user of the              // 138
// server, and it overrides all packages with the same application ID. If any instances of those             // 139
// packages are currently open, they are killed and reset on publish.                                        // 140
//                                                                                                           // 141
// When the dev tool disconnects, the package is automatically unpublished, and any open instances           // 142
// are again killed and refreshed.                                                                           // 143
//                                                                                                           // 144
// Each contains:                                                                                            // 145
//   _id:  The package ID string (as with Packages._id).                                                     // 146
//   appId: The app ID this package is intended to override (as with Packages.appId).                        // 147
//   timestamp:  Time when the package was last updated. If this changes while the package is                // 148
//     published, all running instances are reset. This is used e.g. to reset the app each time              // 149
//     changes are made to the source code.                                                                  // 150
//   manifest:  The app's manifest, as with Packages.manifest.                                               // 151
                                                                                                             // 152
UserActions = new Mongo.Collection("userActions");                                                           // 153
// List of actions that each user has installed which create new grains.  Each app may install               // 154
// some number of actions (usually, one).                                                                    // 155
//                                                                                                           // 156
// Each contains:                                                                                            // 157
//   _id:  random                                                                                            // 158
//   userId:  User who has installed this action.                                                            // 159
//   packageId:  Package used to run this action.                                                            // 160
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 161
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                       // 162
//       that clients can access it without subscribing to the Packages collection.                          // 163
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 164
//       searchability.                                                                                      // 165
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                     // 166
//   title: JSON-encoded LocalizedText title for this action, e.g.                                           // 167
//       `{defaultText: "New Spreadsheet"}`.                                                                 // 168
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.              // 169
//   command:  Manifest.Command to run this action (see package.capnp).                                      // 170
                                                                                                             // 171
Grains = new Mongo.Collection("grains");                                                                     // 172
// Grains belonging to users.                                                                                // 173
//                                                                                                           // 174
// Each contains:                                                                                            // 175
//   _id:  random                                                                                            // 176
//   packageId:  _id of the package of which this grain is an instance.                                      // 177
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 178
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 179
//       searchability.                                                                                      // 180
//   userId: The _id of the account that owns this grain.                                                    // 181
//   identityId: The identity with which the owning account prefers to open this grain.                      // 182
//   title:  Human-readable string title, as chosen by the user.                                             // 183
//   lastUsed:  Date when the grain was last used by a user.                                                 // 184
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                          // 185
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent          // 186
//                   time a session to this grain was opened.                                                // 187
//                                                                                                           // 188
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and           // 189
// web publishing functionality without powerbox support; they will be replaced once the powerbox            // 190
// is implemented.                                                                                           // 191
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and           // 192
//       web publishing. This field is initialized when first requested by the app.                          // 193
                                                                                                             // 194
RoleAssignments = new Mongo.Collection("roleAssignments");                                                   // 195
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges          // 196
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                   // 197
// collection.                                                                                               // 198
                                                                                                             // 199
Contacts = new Mongo.Collection("contacts");                                                                 // 200
// Edges in the social graph.                                                                                // 201
//                                                                                                           // 202
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob               // 203
// will show up in her user-picker UI for actions like share-by-identity.                                    // 204
//                                                                                                           // 205
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of              // 206
// Bob's.                                                                                                    // 207
//                                                                                                           // 208
// Each contains:                                                                                            // 209
//   _id: random                                                                                             // 210
//   ownerId: The accountId of the user account who owns this contact.                                       // 211
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies              // 212
//            the contact to the owner.                                                                      // 213
//   created: Date when this contact was created.                                                            // 214
//   identityId: The `_id` of the user whose contact info this contains.                                     // 215
                                                                                                             // 216
Sessions = new Mongo.Collection("sessions");                                                                 // 217
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.          // 218
//                                                                                                           // 219
// Each contains:                                                                                            // 220
//   _id:  random                                                                                            // 221
//   grainId:  _id of the grain to which this session is connected.                                          // 222
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the           // 223
//       '*' in WILDCARD_HOST.                                                                               // 224
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some              // 225
//       period.                                                                                             // 226
//   userId:  User ID of the user who owns this session.                                                     // 227
//   identityId:  Identity ID of the user who owns this session.                                             // 228
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens            // 229
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),               // 230
//       anonymous users can get access without an API token and so neither userId nor hashedToken           // 231
//       are present.                                                                                        // 232
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                 // 233
//       interaction.                                                                                        // 234
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                         // 235
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated           // 236
//       until newSession is called on the UiView.                                                           // 237
//   permissions: The permissions for the current identity on this UiView. This isn't populated              // 238
//       until newSession is called on the UiView.                                                           // 239
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first         // 240
//       request                                                                                             // 241
                                                                                                             // 242
SignupKeys = new Mongo.Collection("signupKeys");                                                             // 243
// Invite keys which may be used by users to get access to Sandstorm.                                        // 244
//                                                                                                           // 245
// Each contains:                                                                                            // 246
//   _id:  random                                                                                            // 247
//   used:  Boolean indicating whether this key has already been consumed.                                   // 248
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                // 249
//   email: If this key was sent as an email invite, the email address to which it was sent.                 // 250
                                                                                                             // 251
ActivityStats = new Mongo.Collection("activityStats");                                                       // 252
// Contains usage statistics taken on a regular interval. Each entry is a data point.                        // 253
//                                                                                                           // 254
// Each contains:                                                                                            // 255
//   timestamp: Date when measurements were taken.                                                           // 256
//   daily: Contains stats counts pertaining to the last day before the sample time.                         // 257
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                 // 258
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                 // 259
//                                                                                                           // 260
// Each of daily, weekly, and monthly contains:                                                              // 261
//   activeUsers: The number of unique users who have used a grain on the server in the time                 // 262
//       interval. Only counts logged-in users.                                                              // 263
//   demoUsers: Demo users.                                                                                  // 264
//   appDemoUsers: Users that came in through "app demo".                                                    // 265
//   activeGrains: The number of unique grains that have been used in the time interval.                     // 266
//   apps: An object indexed by app ID recording, for each app:                                              // 267
//       owners: Number of unique owners of this app (counting only grains that still exist).                // 268
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting          // 269
//         only grains that still exist).                                                                    // 270
//       grains: Number of active grains of this app (that still exist).                                     // 271
//       deleted: Number of non-demo grains of this app that were deleted.                                   // 272
//       demoed: Number of demo grains created and expired.                                                  // 273
//       appDemoUsers: Number of app demos initiated with this app.                                          // 274
                                                                                                             // 275
DeleteStats = new Mongo.Collection("deleteStats");                                                           // 276
// Contains records of objects that were deleted, for stat-keeping purposes.                                 // 277
//                                                                                                           // 278
// Each contains:                                                                                            // 279
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                   // 280
//   lastActive: Date of the user's or grain's last activity.                                                // 281
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they           // 282
//     arrived to demo. For others, undefined.                                                               // 283
                                                                                                             // 284
FileTokens = new Mongo.Collection("fileTokens");                                                             // 285
// Tokens corresponding to backup files that are currently stored on the server. A user receives             // 286
// a token when they create a backup file (either by uploading it, or by backing up one of their             // 287
// grains) and may use the token to read the file (either to download it, or to restore a new                // 288
// grain from it).                                                                                           // 289
//                                                                                                           // 290
// Each contains:                                                                                            // 291
//   _id:       The unguessable token string.                                                                // 292
//   name:      Suggested filename.                                                                          // 293
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.              // 294
                                                                                                             // 295
ApiTokens = new Mongo.Collection("apiTokens");                                                               // 296
// Access tokens for APIs exported by apps.                                                                  // 297
//                                                                                                           // 298
// Originally API tokens were only used by external users through the HTTP API endpoint. However,            // 299
// now they are also used to implement SturdyRefs, not just held by external users, but also when            // 300
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                 // 301
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and               // 302
// `MainView`) -- the fields of type `Data` are API tokens.                                                  // 303
//                                                                                                           // 304
// Each contains:                                                                                            // 305
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                 // 306
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)            // 307
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.             // 308
//              That is, the UiView's newSession() method will intersect the requested permissions           // 309
//              with this identity's permissions before forwarding on to the underlying app. If              // 310
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is           // 311
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may              // 312
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that           // 313
//              this is NOT the identity against which the `requiredPermissions` parameter of                // 314
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.         // 315
//   accountId: For tokens where `identityId` is set, the `_id` (in the Users table) of the account          // 316
//              that created the token.                                                                      // 317
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded               // 318
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These          // 319
//              permissions will be intersected with those held by `identityId` when the view is             // 320
//              opened.                                                                                      // 321
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as          // 322
//              anonymous rather than as directly associated with `identityId`. This has no effect           // 323
//              on the permissions granted.                                                                  // 324
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by            // 325
//              the app or its supervisor (whereas without this it strictly represents UiView).              // 326
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).            // 327
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                   // 328
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                     // 329
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,          // 330
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object          // 331
//              containing exactly one of the following fields:                                              // 332
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                          // 333
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                 // 334
//                           `Notifications` collection.                                                     // 335
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this            // 336
//                  will be moved out of the frontend and into the backend, but we'll migrate the            // 337
//                  database when that happens. This field contains the boolean true to signify that         // 338
//                  it has been set.                                                                         // 339
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                           // 340
//   parentToken: If present, then this token represents exactly the capability represented by               // 341
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated          // 342
//              by `roleAssignment` (if present). To facilitate permissions computations, if the             // 343
//              capability is a UiView, then `grainId` is set to the backing grain, `identityId`             // 344
//              is set to the identity that shared the view, and `accountId` is set to the account           // 345
//              that shared the view. Neither `objectId` nor `frontendRef` is present when                   // 346
//              `parentToken` is present.                                                                    // 347
//   petname:   Human-readable label for this access token, useful for identifying tokens for                // 348
//              revocation. This should be displayed when visualizing incoming capabilities to               // 349
//              the grain identified by `grainId`.                                                           // 350
//   created:   Date when this token was created.                                                            // 351
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may          // 352
//              become un-revoked in the future.                                                             // 353
//   expires:   Optional expiration Date. If undefined, the token does not expire.                           // 354
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)                // 355
//              as passed to the `save()` call that created this token. If not present, treat                // 356
//              as `webkey` (the default for `ApiTokenOwner`).                                               // 357
//   expiresIfUnused:                                                                                        // 358
//              Optional Date after which the token, if it has not been used yet, expires.                   // 359
//              This field should be cleared on a token's first use.                                         // 360
//   requirements: List of conditions which must hold for this token to be considered valid.                 // 361
//              Semantically, this list specifies the powers which were *used* to originally                 // 362
//              create the token. If any condition in the list becomes untrue, then the token must           // 363
//              be considered revoked, and all live refs and sturdy refs obtained transitively               // 364
//              through it must also become revoked. Each item is the JSON serialization of the              // 365
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                               // 366
//                                                                                                           // 367
// It is important to note that a token's owner and provider are independent from each other. To             // 368
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:          // 369
//                                                                                                           // 370
// struct ApiToken {                                                                                         // 371
//   owner :ApiTokenOwner;                                                                                   // 372
//   provider :union {                                                                                       // 373
//     grain :group {                                                                                        // 374
//       grainId :Text;                                                                                      // 375
//       union {                                                                                             // 376
//         uiView :group {                                                                                   // 377
//           identityId :Text;                                                                               // 378
//           roleAssignment :RoleAssignment;                                                                 // 379
//           forSharing :Bool;                                                                               // 380
//         }                                                                                                 // 381
//         objectId :SupervisorObjectId;                                                                     // 382
//       }                                                                                                   // 383
//     }                                                                                                     // 384
//     frontendRef :union {                                                                                  // 385
//        notificationHandle :Text;                                                                          // 386
//        ipNetwork :Bool;                                                                                   // 387
//        ipInterface :Bool;                                                                                 // 388
//     }                                                                                                     // 389
//     child :group {                                                                                        // 390
//       parentToken :Text;                                                                                  // 391
//       union {                                                                                             // 392
//         uiView :group {                                                                                   // 393
//           grainId :Text;                                                                                  // 394
//           identityId :Text;                                                                               // 395
//           roleAssignment :RoleAssignment = (allAccess = ());                                              // 396
//         }                                                                                                 // 397
//         other :Void;                                                                                      // 398
//       }                                                                                                   // 399
//     }                                                                                                     // 400
//   }                                                                                                       // 401
//   requirements: List(Supervisor.MembraneRequirement);                                                     // 402
//   ...                                                                                                     // 403
// }                                                                                                         // 404
                                                                                                             // 405
Notifications = new Mongo.Collection("notifications");                                                       // 406
// Notifications for a user.                                                                                 // 407
//                                                                                                           // 408
// Each contains:                                                                                            // 409
//   _id:          random                                                                                    // 410
//   grainId:      The grain originating this notification, if any.                                          // 411
//   userId:       The user receiving the notification.                                                      // 412
//   text:         The JSON-ified LocalizedText to display in the notification.                              // 413
//   isUnread:     Boolean indicating if this notification is unread.                                        // 414
//   timestamp:    Date when this notification was last updated                                              // 415
//   ongoing:      If present, this is an ongoing notification, and this field contains an                   // 416
//                 ApiToken referencing the `OngoingNotification` capability.                                // 417
//   admin:        If present, this is a notification intended for an admin.                                 // 418
//     action:     If present, this is a (string) link that the notification should direct the               // 419
//                 admin to.                                                                                 // 420
//     type:       The type of notification (currently only "reportStats").                                  // 421
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds           // 422
//                 as keys.                                                                                  // 423
//     $appId:     The appId that has an outstanding update.                                                 // 424
//       packageId: The packageId that it will update to.                                                    // 425
//       name: The name of the app. (appTitle from package.manifest)                                         // 426
//       version: The app's version number. (appVersion from package.manifest)                               // 427
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)
//   referral:     If this boolean field is true, then treat this notification as a referral                 // 429
//                 notification. This causes text to be ignored, since we need custom logic.                 // 430
                                                                                                             // 431
StatsTokens = new Mongo.Collection("statsTokens");                                                           // 432
// Access tokens for the Stats collection                                                                    // 433
//                                                                                                           // 434
// These tokens are used for accessing the ActivityStats collection remotely                                 // 435
// (ie. from a dashboard webapp)                                                                             // 436
//                                                                                                           // 437
// Each contains:                                                                                            // 438
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                        // 439
                                                                                                             // 440
Misc = new Mongo.Collection("misc");                                                                         // 441
// Miscellaneous configuration and other settings                                                            // 442
//                                                                                                           // 443
// This table is currently only used for persisting BASE_URL from one session to the next,                   // 444
// but in general any miscellaneous settings should go in here                                               // 445
//                                                                                                           // 446
// Each contains:                                                                                            // 447
//   _id:       The name of the setting. eg. "BASE_URL"                                                      // 448
//   value:     The value of the setting.                                                                    // 449
                                                                                                             // 450
Settings = new Mongo.Collection("settings");                                                                 // 451
// Settings for this Sandstorm instance go here. They are configured through the adminSettings               // 452
// route. This collection differs from misc in that any admin user can update it through the admin           // 453
// interface.                                                                                                // 454
//                                                                                                           // 455
// Each contains:                                                                                            // 456
//   _id:       The name of the setting. eg. "MAIL_URL"                                                      // 457
//   value:     The value of the setting.                                                                    // 458
//   automaticallyReset: Sometimes the server needs to automatically reset a setting. When it does           // 459
//                       so, it will also write an object to this field indicating why the reset was         // 460
//                       needed. That object can have the following variants:                                // 461
//       baseUrlChangedFrom: The reset was due to BASE_URL changing. This field contains a string            // 462
//                           with the old BASE_URL.                                                          // 463
//                                                                                                           // 464
//   potentially other fields that are unique to the setting                                                 // 465
                                                                                                             // 466
Migrations = new Mongo.Collection("migrations");                                                             // 467
// This table tracks which migrations we have applied to this instance.                                      // 468
// It contains a single entry:                                                                               // 469
//   _id:       "migrations_applied"                                                                         // 470
//   value:     The number of migrations this instance has successfully completed.                           // 471
                                                                                                             // 472
StaticAssets = new Mongo.Collection("staticAssets");                                                         // 473
// Collection of static assets served up from the Sandstorm server's "static" host. We only                  // 474
// support relatively small assets: under 1MB each.                                                          // 475
//                                                                                                           // 476
// Each contains:                                                                                            // 477
//   _id:       Random ID; will be used in the URL.                                                          // 478
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                  // 479
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                    // 480
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                          // 481
//   content:   The asset content (byte buffer).                                                             // 482
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't          // 483
//       have transactions, this needs to bias towards over-counting; a backup GC could be used              // 484
//       to catch leaked assets, although it's probably not a big deal in practice.                          // 485
                                                                                                             // 486
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                               // 487
// Collection of tokens representing a single-use permission to upload an asset, such as a new               // 488
// profile picture.                                                                                          // 489
//                                                                                                           // 490
// Each contains:                                                                                            // 491
//   _id:       Random ID.                                                                                   // 492
//   purpose:   Contains one of the following, indicating how the asset is to be used:                       // 493
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                // 494
//           userId: User whose picture shall be replaced.                                                   // 495
//           identityId: Which of the user's identities shall be updated.                                    // 496
//   expires:   Time when this token will go away if unused.                                                 // 497
                                                                                                             // 498
Plans = new Mongo.Collection("plans");                                                                       // 499
// Subscription plans, which determine quota.                                                                // 500
//                                                                                                           // 501
// Each contains:                                                                                            // 502
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                      // 503
//   storage: Number of bytes this user is allowed to store.                                                 // 504
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                // 505
//   computeLabel: Label to display to the user describing this plan's compute units.                        // 506
//   grains: Total number of grains this user can create (often `Infinity`).                                 // 507
//   price: Price per month in US cents.                                                                     // 508
//   hidden: If true, a user cannot switch to this plan, but some users may be on it and are                 // 509
//       allowed to switch away.                                                                             // 510
//   title: Title from display purposes. If missing, default to capitalizing _id.                            // 511
                                                                                                             // 512
AppIndex = new Mongo.Collection("appIndex");                                                                 // 513
// A mirror of the data from the App Market index                                                            // 514
//                                                                                                           // 515
// Each contains:                                                                                            // 516
//   _id: the appId of the app                                                                               // 517
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket          // 518
                                                                                                             // 519
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                   // 520
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package           // 521
// by that user is installed, as well as if the keybase profile is requested and not already                 // 522
// present for some reason.                                                                                  // 523
//                                                                                                           // 524
// Each contains:                                                                                            // 525
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                         // 526
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                          // 527
//   handle: Keybase handle.                                                                                 // 528
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase            // 529
//     docs for details. We also add a boolean "status" field to each proof indicating whether               // 530
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                // 531
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it           // 532
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the           // 533
//     server is currently actively checking this proof). Note that if a check fails due to network          // 534
//     errors, the status goes back to "unverified".                                                         // 535
//                                                                                                           // 536
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                // 537
//       for now and we just trust Keybase.                                                                  // 538
                                                                                                             // 539
if (Meteor.isServer) {                                                                                       // 540
  Meteor.publish("credentials", function () {                                                                // 541
    // Data needed for isSignedUp() and isAdmin() to work.                                                   // 542
                                                                                                             // 543
    if (this.userId) {                                                                                       // 544
      return [                                                                                               // 545
        Meteor.users.find({_id: this.userId},                                                                // 546
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                                 // 547
                      plan: 1, hasCompletedSignup: 1, experiments: 1}}),                                     // 548
        Plans.find()                                                                                         // 549
      ];                                                                                                     // 550
    } else {                                                                                                 // 551
      return [];                                                                                             // 552
    }                                                                                                        // 553
  });                                                                                                        // 554
}                                                                                                            // 555
                                                                                                             // 556
isDemoUser = function() {                                                                                    // 557
  // Returns true if this is a demo user.                                                                    // 558
                                                                                                             // 559
  var user = Meteor.user();                                                                                  // 560
  if (user && user.expires) {                                                                                // 561
    return true;                                                                                             // 562
  } else {                                                                                                   // 563
    return false;                                                                                            // 564
  }                                                                                                          // 565
}                                                                                                            // 566
                                                                                                             // 567
isSignedUp = function() {                                                                                    // 568
  // Returns true if the user has presented an invite key.                                                   // 569
                                                                                                             // 570
  var user = Meteor.user();                                                                                  // 571
                                                                                                             // 572
  if (!user) return false;  // not signed in                                                                 // 573
                                                                                                             // 574
  if (!user.loginIdentities) return false;  // not an account                                                // 575
                                                                                                             // 576
  if (user.expires) return false;  // demo user.                                                             // 577
                                                                                                             // 578
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 579
                                                                                                             // 580
  if (user.signupKey) return true;  // user is invited                                                       // 581
                                                                                                             // 582
  return false;                                                                                              // 583
}                                                                                                            // 584
                                                                                                             // 585
isSignedUpOrDemo = function () {                                                                             // 586
  var user = Meteor.user();                                                                                  // 587
                                                                                                             // 588
  if (!user) return false;  // not signed in                                                                 // 589
                                                                                                             // 590
  if (!user.loginIdentities) return false;  // not an account                                                // 591
                                                                                                             // 592
  if (user.expires) return true;  // demo user.                                                              // 593
                                                                                                             // 594
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 595
                                                                                                             // 596
  if (user.signupKey) return true;  // user is invited                                                       // 597
                                                                                                             // 598
  return false;                                                                                              // 599
}                                                                                                            // 600
                                                                                                             // 601
var calculateReferralBonus = function(accountId, plan) {                                                     // 602
  // This function returns an object of the form:                                                            // 603
  //                                                                                                         // 604
  // - {grains: 0, storage: 0}                                                                               // 605
  //                                                                                                         // 606
  // which are extra resources this account gets as part of participating in the referral                    // 607
  // program. (Storage is measured in bytes, as usual for plans.)                                            // 608
                                                                                                             // 609
                                                                                                             // 610
  // Authorization note: Only call this if accountId is the current user!                                    // 611
  var isPaid = (plan && plan !== "free");                                                                    // 612
                                                                                                             // 613
  successfulReferralsCount = countReferrals(accountId);                                                      // 614
  if (isPaid) {                                                                                              // 615
    var maxPaidStorageBonus = 30 * 1e9;                                                                      // 616
    return {grains: 0,                                                                                       // 617
            storage: Math.min(                                                                               // 618
              successfulReferralsCount * 2 * 1e9,                                                            // 619
              maxPaidStorageBonus)};                                                                         // 620
  } else {                                                                                                   // 621
    var maxFreeStorageBonus = 2 * 1e9;                                                                       // 622
    var bonus = {                                                                                            // 623
      storage: Math.min(                                                                                     // 624
        successfulReferralsCount * 50 * 1e6,                                                                 // 625
        maxFreeStorageBonus)                                                                                 // 626
    };                                                                                                       // 627
    if (successfulReferralsCount > 0) {                                                                      // 628
      bonus.grains = Infinity;                                                                               // 629
    } else {                                                                                                 // 630
      bonus.grains = 0;                                                                                      // 631
    }                                                                                                        // 632
    return bonus;                                                                                            // 633
  }                                                                                                          // 634
}                                                                                                            // 635
                                                                                                             // 636
var countReferrals = function (accountId) {                                                                  // 637
  var referredIdentityIds = Meteor.users.findOne(                                                            // 638
    {_id: accountId},                                                                                        // 639
    {fields: {referredIdentityIds: 1}}).referredIdentityIds;                                                 // 640
  return (referredIdentityIds && referredIdentityIds.length || 0);                                           // 641
}                                                                                                            // 642
                                                                                                             // 643
getUserQuota = function (user) {                                                                             // 644
  var plan = Plans.findOne(user.plan || "free");                                                             // 645
  var referralBonus = calculateReferralBonus(user._id, plan);                                                // 646
  var userQuota = {};                                                                                        // 647
  userQuota.storage = (plan.storage + referralBonus.storage);                                                // 648
  userQuota.grains = (plan.grains + referralBonus.grains);                                                   // 649
  return userQuota;                                                                                          // 650
}                                                                                                            // 651
                                                                                                             // 652
isUserOverQuota = function (user) {                                                                          // 653
  // Return false if user has quota space remaining, true if it is full. When this returns true,             // 654
  // we will not allow the user to create new grains, though they may be able to open existing ones          // 655
  // which may still increase their storage usage.                                                           // 656
  //                                                                                                         // 657
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 658
                                                                                                             // 659
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 660
                                                                                                             // 661
  var plan = getUserQuota(user);                                                                             // 662
  if (plan.grains < Infinity) {                                                                              // 663
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();                   // 664
    if (count >= plan.grains) return "outOfGrains";                                                          // 665
  }                                                                                                          // 666
                                                                                                             // 667
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                   // 668
}                                                                                                            // 669
                                                                                                             // 670
isUserExcessivelyOverQuota = function (user) {                                                               // 671
  // Return true if user is so far over quota that we should prevent their existing grains from              // 672
  // running at all.                                                                                         // 673
  //                                                                                                         // 674
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 675
                                                                                                             // 676
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 677
                                                                                                             // 678
  var quota = getUserQuota(user);                                                                            // 679
                                                                                                             // 680
  // quota.grains = Infinity means unlimited grains. IEEE754 defines Infinity == Infinity.                   // 681
  if (quota.grains < Infinity) {                                                                             // 682
    var count = Grains.find({userId: user._id}, {fields: {}, limit: quota.grains * 2}).count();              // 683
    if (count >= quota.grains * 2) return "outOfGrains";                                                     // 684
  }                                                                                                          // 685
                                                                                                             // 686
  return quota && user.storageUsage && user.storageUsage >= quota.storage * 1.2 && "outOfStorage";           // 687
}                                                                                                            // 688
                                                                                                             // 689
isAdmin = function() {                                                                                       // 690
  // Returns true if the user is the administrator.                                                          // 691
                                                                                                             // 692
  var user = Meteor.user();                                                                                  // 693
  if (user && user.isAdmin) {                                                                                // 694
    return true;                                                                                             // 695
  } else {                                                                                                   // 696
    return false;                                                                                            // 697
  }                                                                                                          // 698
}                                                                                                            // 699
                                                                                                             // 700
isAdminById = function(id) {                                                                                 // 701
  // Returns true if the user's id is the administrator.                                                     // 702
                                                                                                             // 703
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                         // 704
  if (user && user.isAdmin) {                                                                                // 705
    return true;                                                                                             // 706
  } else {                                                                                                   // 707
    return false;                                                                                            // 708
  }                                                                                                          // 709
}                                                                                                            // 710
                                                                                                             // 711
findAdminUserForToken = function (token) {                                                                   // 712
  if (!token.requirements) {                                                                                 // 713
    return;                                                                                                  // 714
  }                                                                                                          // 715
  var requirements = token.requirements.filter(function (requirement) {                                      // 716
    return "userIsAdmin" in requirement;                                                                     // 717
  });                                                                                                        // 718
                                                                                                             // 719
  if (requirements.length > 1) {                                                                             // 720
    return;                                                                                                  // 721
  }                                                                                                          // 722
  if (requirements.length === 0) {                                                                           // 723
    return;                                                                                                  // 724
  }                                                                                                          // 725
  return requirements[0].userIsAdmin;                                                                        // 726
};                                                                                                           // 727
                                                                                                             // 728
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                             // 729
                                                                                                             // 730
if (wildcardHost.length != 2) {                                                                              // 731
  throw new Error("Wildcard host must contain exactly one asterisk.");                                       // 732
}                                                                                                            // 733
                                                                                                             // 734
matchWildcardHost = function(host) {                                                                         // 735
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                 // 736
                                                                                                             // 737
  var prefix = wildcardHost[0];                                                                              // 738
  var suffix = wildcardHost[1];                                                                              // 739
                                                                                                             // 740
  // We remove everything after the first ":" character so that our                                          // 741
  // comparison logic ignores port numbers.                                                                  // 742
  suffix = suffix.split(":")[0];                                                                             // 743
  host = host.split(":")[0];                                                                                 // 744
                                                                                                             // 745
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                                    // 746
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                           // 747
      host.length >= prefix.length + suffix.length) {                                                        // 748
    var id = host.slice(prefix.length, -suffix.length);                                                      // 749
    if (id.match(/^[a-z0-9]*$/)) {                                                                           // 750
      return id;                                                                                             // 751
    }                                                                                                        // 752
  }                                                                                                          // 753
                                                                                                             // 754
  return null;                                                                                               // 755
}                                                                                                            // 756
                                                                                                             // 757
makeWildcardHost = function (id) {                                                                           // 758
  return wildcardHost[0] + id + wildcardHost[1];                                                             // 759
}                                                                                                            // 760
                                                                                                             // 761
if (Meteor.isServer) {                                                                                       // 762
  var Url = Npm.require("url");                                                                              // 763
  getWildcardOrigin = function () {                                                                          // 764
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're              // 765
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),            // 766
    // an astrisk is only allowed as the first character and must be followed by a period. So we need        // 767
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can           // 768
    // really do. We also add the protocol to the front (again, that's what CSP wants).                      // 769
                                                                                                             // 770
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.         // 771
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                 // 772
                                                                                                             // 773
    var dotPos = wildcardHost[1].indexOf(".");                                                               // 774
    if (dotPos < 0) {                                                                                        // 775
      return protocol + "//*";                                                                               // 776
    } else {                                                                                                 // 777
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                               // 778
    }                                                                                                        // 779
  }                                                                                                          // 780
}                                                                                                            // 781
                                                                                                             // 782
allowDevAccounts = function () {                                                                             // 783
  var setting = Settings.findOne({_id: "devAccounts"});                                                      // 784
  if (setting) {                                                                                             // 785
    return setting.value;                                                                                    // 786
  } else {                                                                                                   // 787
    return Meteor.settings && Meteor.settings.public &&                                                      // 788
           Meteor.settings.public.allowDevAccounts;                                                          // 789
  }                                                                                                          // 790
};                                                                                                           // 791
                                                                                                             // 792
sendReferralProgramNotification = function (userId) {                                                        // 793
  Notifications.upsert({                                                                                     // 794
    userId: userId,                                                                                          // 795
    referral: true,                                                                                          // 796
  }, {                                                                                                       // 797
    userId: userId,                                                                                          // 798
    referral: true,                                                                                          // 799
    timestamp: new Date(),                                                                                   // 800
    isUnread: true,                                                                                          // 801
  });                                                                                                        // 802
};                                                                                                           // 803
                                                                                                             // 804
roleAssignmentPattern = {                                                                                    // 805
  none : Match.Optional(null),                                                                               // 806
  allAccess: Match.Optional(null),                                                                           // 807
  roleId: Match.Optional(Match.Integer),                                                                     // 808
  addPermissions: Match.Optional([Boolean]),                                                                 // 809
  removePermissions: Match.Optional([Boolean]),                                                              // 810
};                                                                                                           // 811
                                                                                                             // 812
SandstormDb = function () {                                                                                  // 813
  this.collections = {                                                                                       // 814
    // Direct access to underlying collections. DEPRECATED.                                                  // 815
    //                                                                                                       // 816
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove            // 817
    //   direct access to the collections.                                                                   // 818
                                                                                                             // 819
    packages: Packages,                                                                                      // 820
    devPackages: DevPackages,                                                                                // 821
    userActions: UserActions,                                                                                // 822
    grains: Grains,                                                                                          // 823
    contacts: Contacts,                                                                                      // 824
    sessions: Sessions,                                                                                      // 825
    signupKeys: SignupKeys,                                                                                  // 826
    activityStats: ActivityStats,                                                                            // 827
    deleteStats: DeleteStats,                                                                                // 828
    fileTokens: FileTokens,                                                                                  // 829
    apiTokens: ApiTokens,                                                                                    // 830
    notifications: Notifications,                                                                            // 831
    statsTokens: StatsTokens,                                                                                // 832
    misc: Misc,                                                                                              // 833
    settings: Settings,                                                                                      // 834
    appIndex: AppIndex,                                                                                      // 835
    keybaseProfiles: KeybaseProfiles,                                                                        // 836
                                                                                                             // 837
    // Intentionally omitted:                                                                                // 838
    // - Migrations, since it's used only within this package.                                               // 839
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.          // 840
  };                                                                                                         // 841
};                                                                                                           // 842
                                                                                                             // 843
// TODO(cleanup): These methods should not be defined freestanding and should use collection                 // 844
//   objects created in SandstormDb's constructor rather than globals.                                       // 845
                                                                                                             // 846
_.extend(SandstormDb.prototype, {                                                                            // 847
  isDemoUser: isDemoUser,                                                                                    // 848
  isSignedUp: isSignedUp,                                                                                    // 849
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                        // 850
  isUserOverQuota: isUserOverQuota,                                                                          // 851
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                    // 852
  isAdmin: isAdmin,                                                                                          // 853
  isAdminById: isAdminById,                                                                                  // 854
  findAdminUserForToken: findAdminUserForToken,                                                              // 855
  matchWildcardHost: matchWildcardHost,                                                                      // 856
  makeWildcardHost: makeWildcardHost,                                                                        // 857
  allowDevAccounts: allowDevAccounts,                                                                        // 858
  roleAssignmentPattern: roleAssignmentPattern,                                                              // 859
});                                                                                                          // 860
                                                                                                             // 861
if (Meteor.isServer) {                                                                                       // 862
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                               // 863
}                                                                                                            // 864
                                                                                                             // 865
// =======================================================================================                   // 866
// Below this point are newly-written or refactored functions.                                               // 867
                                                                                                             // 868
_.extend(SandstormDb.prototype, {                                                                            // 869
  getUser: function getUser (userId) {                                                                       // 870
    check(userId, Match.OneOf(String, undefined, null));                                                     // 871
    if (userId) {                                                                                            // 872
      return Meteor.users.findOne(userId);                                                                   // 873
    }                                                                                                        // 874
  },                                                                                                         // 875
                                                                                                             // 876
  getIdentity: function getIdentity (identityId) {                                                           // 877
    check(identityId, String);                                                                               // 878
    var identity = Meteor.users.findOne({_id: identityId});                                                  // 879
    if (identity) {                                                                                          // 880
      SandstormDb.fillInProfileDefaults(identity);                                                           // 881
      SandstormDb.fillInIntrinsicName(identity);                                                             // 882
      SandstormDb.fillInPictureUrl(identity);                                                                // 883
      return identity;                                                                                       // 884
    }                                                                                                        // 885
  },                                                                                                         // 886
                                                                                                             // 887
  userHasIdentity: function (userId, identityId) {                                                           // 888
    check(userId, String);                                                                                   // 889
    check(identityId, String);                                                                               // 890
                                                                                                             // 891
    if (userId === identityId) return true;                                                                  // 892
                                                                                                             // 893
    var user = Meteor.users.findOne(userId);                                                                 // 894
    return SandstormDb.getUserIdentityIds(user).indexOf(identityId) != -1;                                   // 895
  },                                                                                                         // 896
                                                                                                             // 897
  userGrains: function userGrains (userId) {                                                                 // 898
    check(userId, Match.OneOf(String, undefined, null));                                                     // 899
    return this.collections.grains.find({userId: userId});                                                   // 900
  },                                                                                                         // 901
                                                                                                             // 902
  currentUserGrains: function currentUserGrains () {                                                         // 903
    return this.userGrains(Meteor.userId());                                                                 // 904
  },                                                                                                         // 905
                                                                                                             // 906
  getGrain: function getGrain (grainId) {                                                                    // 907
    check(grainId, String);                                                                                  // 908
    return this.collections.grains.findOne(grainId);                                                         // 909
  },                                                                                                         // 910
                                                                                                             // 911
  userApiTokens: function userApiTokens (userId) {                                                           // 912
    check(userId, Match.OneOf(String, undefined, null));                                                     // 913
    var identityIds = SandstormDb.getUserIdentityIds(this.getUser(userId));                                  // 914
    return this.collections.apiTokens.find({'owner.user.identityId': {$in: identityIds}});                   // 915
  },                                                                                                         // 916
                                                                                                             // 917
  currentUserApiTokens: function currentUserApiTokens () {                                                   // 918
    return this.userApiTokens(Meteor.userId());                                                              // 919
  },                                                                                                         // 920
                                                                                                             // 921
  userActions: function userActions (user) {                                                                 // 922
    return this.collections.userActions.find({userId: user});                                                // 923
  },                                                                                                         // 924
                                                                                                             // 925
  currentUserActions: function currentUserActions () {                                                       // 926
    return this.userActions(Meteor.userId());                                                                // 927
  },                                                                                                         // 928
                                                                                                             // 929
  iconSrcForPackage: function iconSrcForPackage (pkg, usage) {                                               // 930
    return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                         // 931
  },                                                                                                         // 932
                                                                                                             // 933
  getPlan: function (id) {                                                                                   // 934
    check(id, String);                                                                                       // 935
    var plan = Plans.findOne(id);                                                                            // 936
    if (!plan) {                                                                                             // 937
      throw new Error("no such plan: " + id);                                                                // 938
    }                                                                                                        // 939
    return plan;                                                                                             // 940
  },                                                                                                         // 941
                                                                                                             // 942
  listPlans: function () {                                                                                   // 943
    return Plans.find({}, {sort: {price: 1}});                                                               // 944
  },                                                                                                         // 945
                                                                                                             // 946
  getMyPlan: function () {                                                                                   // 947
    var user = Meteor.user();                                                                                // 948
    return user && Plans.findOne(user.plan || "free");                                                       // 949
  },                                                                                                         // 950
                                                                                                             // 951
  getMyReferralBonus: function(user) {                                                                       // 952
    // This function is called from the server and from the client, similar to getMyPlan().                  // 953
    //                                                                                                       // 954
    // When called from the server, calculate the user's actual referral bonus. We use this                  // 955
    // elsewhere to store a value in user.pseudoReferralBonus.                                               // 956
    //                                                                                                       // 957
    // When called from the client, return the value of user.pseudoReferralBonus (if it exists).             // 958
    if (Meteor.isClient) {                                                                                   // 959
      // If called on the client side, always use the currently logged-in user.                              // 960
      user = Meteor.user();                                                                                  // 961
                                                                                                             // 962
      if (user && user.pseudoReferralBonus) {                                                                // 963
        return user.pseudoReferralBonus;                                                                     // 964
      }                                                                                                      // 965
      // If we get to this, the subscriptions haven't arrived yet.                                           // 966
      var noBonus = {grains: 0, storage: 0};                                                                 // 967
      return noBonus;                                                                                        // 968
    }                                                                                                        // 969
    if (Meteor.isServer) {                                                                                   // 970
      var x = calculateReferralBonus(user._id, user.plan);                                                   // 971
      return x;                                                                                              // 972
    }                                                                                                        // 973
  },                                                                                                         // 974
                                                                                                             // 975
  getMyUsage: function (user) {                                                                              // 976
    user = user || Meteor.user();                                                                            // 977
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                     // 978
      if (Meteor.isClient) {                                                                                 // 979
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly              // 980
        // not reactive.                                                                                     // 981
        return user.pseudoUsage;                                                                             // 982
      } else {                                                                                               // 983
        return {                                                                                             // 984
          grains: Grains.find({userId: user._id}).count(),                                                   // 985
          storage: user.storageUsage || 0,                                                                   // 986
          compute: 0   // not tracked yet                                                                    // 987
        };                                                                                                   // 988
      }                                                                                                      // 989
    } else {                                                                                                 // 990
      return {grains: 0, storage: 0, compute: 0};                                                            // 991
    }                                                                                                        // 992
  },                                                                                                         // 993
                                                                                                             // 994
  isUninvitedFreeUser: function () {                                                                         // 995
    if (!Meteor.settings.public.allowUninvited) return false;                                                // 996
                                                                                                             // 997
    var user = Meteor.user();                                                                                // 998
    return user && !user.expires && (!user.plan || user.plan === "free");                                    // 999
  },                                                                                                         // 1000
                                                                                                             // 1001
  getSetting: function (name) {                                                                              // 1002
    var setting = Settings.findOne(name);                                                                    // 1003
    return setting && setting.value;                                                                         // 1004
  },                                                                                                         // 1005
                                                                                                             // 1006
  addUserActions: function(packageId) {                                                                      // 1007
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.             // 1008
    var pack = Packages.findOne(packageId);                                                                  // 1009
    if (pack) {                                                                                              // 1010
      // Remove old versions.                                                                                // 1011
      UserActions.find({userId: Meteor.userId(), appId: pack.appId})                                         // 1012
          .forEach(function (action) {                                                                       // 1013
        UserActions.remove(action._id);                                                                      // 1014
      });                                                                                                    // 1015
                                                                                                             // 1016
      // Install new.                                                                                        // 1017
      var actions = pack.manifest.actions;                                                                   // 1018
      for (i in actions) {                                                                                   // 1019
        var action = actions[i];                                                                             // 1020
        if ("none" in action.input) {                                                                        // 1021
          var userAction = {                                                                                 // 1022
            userId: Meteor.userId(),                                                                         // 1023
            packageId: pack._id,                                                                             // 1024
            appId: pack.appId,                                                                               // 1025
            appTitle: pack.manifest.appTitle,                                                                // 1026
            appMarketingVersion: pack.manifest.appMarketingVersion,                                          // 1027
            appVersion: pack.manifest.appVersion,                                                            // 1028
            title: action.title,                                                                             // 1029
            nounPhrase: action.nounPhrase,                                                                   // 1030
            command: action.command                                                                          // 1031
          };                                                                                                 // 1032
          UserActions.insert(userAction);                                                                    // 1033
        } else {                                                                                             // 1034
          // TODO(someday):  Implement actions with capability inputs.                                       // 1035
        }                                                                                                    // 1036
      }                                                                                                      // 1037
                                                                                                             // 1038
      Meteor.call("deleteUnusedPackages", pack.appId);                                                       // 1039
    }                                                                                                        // 1040
  },                                                                                                         // 1041
                                                                                                             // 1042
  sendAdminNotification: function (message, link) {                                                          // 1043
    Meteor.users.find({isAdmin: true}, {fields: {_id: 1}}).forEach(function (user) {                         // 1044
      Notifications.insert({                                                                                 // 1045
        admin: {                                                                                             // 1046
          action: link,                                                                                      // 1047
          type: "reportStats",                                                                               // 1048
        },                                                                                                   // 1049
        userId: user._id,                                                                                    // 1050
        text: {defaultText: message},                                                                        // 1051
        timestamp: new Date(),                                                                               // 1052
        isUnread: true,                                                                                      // 1053
      });                                                                                                    // 1054
    });                                                                                                      // 1055
  },                                                                                                         // 1056
                                                                                                             // 1057
  getKeybaseProfile: function (keyFingerprint) {                                                             // 1058
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                   // 1059
  },                                                                                                         // 1060
                                                                                                             // 1061
  getServerTitle: function () {                                                                              // 1062
    return Settings.findOne({_id: "serverTitle"}).value;                                                     // 1063
  },                                                                                                         // 1064
                                                                                                             // 1065
  getReturnAddress: function () {                                                                            // 1066
    return Settings.findOne({_id: "returnAddress"}).value;                                                   // 1067
  }                                                                                                          // 1068
});                                                                                                          // 1069
                                                                                                             // 1070
var appNameFromPackage = function(packageObj) {                                                              // 1071
  // This function takes a Package object from Mongo and returns an                                          // 1072
  // app title.                                                                                              // 1073
  var manifest = packageObj.manifest;                                                                        // 1074
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                     // 1075
  var action = manifest.actions[0];                                                                          // 1076
  appName = (manifest.appTitle && manifest.appTitle.defaultText) ||                                          // 1077
    appNameFromActionName(action.title.defaultText);                                                         // 1078
  return appName;                                                                                            // 1079
};                                                                                                           // 1080
                                                                                                             // 1081
var appNameFromActionName = function(name) {                                                                 // 1082
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                     // 1083
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                        // 1084
  //   derive the app title from the action title.                                                           // 1085
  // TODO(cleanup): Get rid of this once apps have real titles.                                              // 1086
  if (!name) {                                                                                               // 1087
    return "(unnamed)";                                                                                      // 1088
  }                                                                                                          // 1089
  if (name.lastIndexOf("New ", 0) === 0) {                                                                   // 1090
    name = name.slice(4);                                                                                    // 1091
  }                                                                                                          // 1092
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                             // 1093
    name = "Hacker CMS";                                                                                     // 1094
  } else {                                                                                                   // 1095
    var space = name.indexOf(" ");                                                                           // 1096
    if (space > 0) {                                                                                         // 1097
      name = name.slice(0, space);                                                                           // 1098
    }                                                                                                        // 1099
  }                                                                                                          // 1100
  return name;                                                                                               // 1101
};                                                                                                           // 1102
                                                                                                             // 1103
var appShortDescriptionFromPackage = function (pkg) {                                                        // 1104
  return pkg && pkg.manifest && pkg.manifest.metadata &&                                                     // 1105
         pkg.manifest.metadata.shortDescription &&                                                           // 1106
         pkg.manifest.metadata.shortDescription.defaultText;                                                 // 1107
};                                                                                                           // 1108
                                                                                                             // 1109
var nounPhraseForActionAndAppTitle = function(action, appTitle) {                                            // 1110
  // A hack to deal with legacy apps not including fields in their manifests.                                // 1111
  // I look forward to the day I can remove most of this code.                                               // 1112
  // Attempt to figure out the appropriate noun that this action will create.                                // 1113
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                  // 1114
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                               // 1115
  // Otherwise, try to guess one from the structure of the action title field                                // 1116
  if (action.title && action.title.defaultText) {                                                            // 1117
    var text = action.title.defaultText;                                                                     // 1118
    // Strip a leading "New "                                                                                // 1119
    if (text.lastIndexOf("New ", 0) === 0) {                                                                 // 1120
      var candidate = text.slice(4);                                                                         // 1121
      // Strip a leading appname too, if provided                                                            // 1122
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                        // 1123
        var newCandidate = candidate.slice(appTitle.length);                                                 // 1124
        // Unless that leaves you with no noun, in which case, use "grain"                                   // 1125
        if (newCandidate.length > 0) {                                                                       // 1126
          return newCandidate.toLowerCase();                                                                 // 1127
        } else {                                                                                             // 1128
          return "grain";                                                                                    // 1129
        }                                                                                                    // 1130
      }                                                                                                      // 1131
      return candidate.toLowerCase();                                                                        // 1132
    }                                                                                                        // 1133
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates              // 1134
    // the package soon.                                                                                     // 1135
    return text;                                                                                             // 1136
  } else {                                                                                                   // 1137
    return "grain";                                                                                          // 1138
  }                                                                                                          // 1139
};                                                                                                           // 1140
                                                                                                             // 1141
// Static methods on SandstormDb that don't need an instance.                                                // 1142
// Largely things that deal with backwards-compatibility.                                                    // 1143
_.extend(SandstormDb, {                                                                                      // 1144
  appNameFromActionName: appNameFromActionName,                                                              // 1145
  appNameFromPackage: appNameFromPackage,                                                                    // 1146
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                            // 1147
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle,                                            // 1148
});                                                                                                          // 1149
                                                                                                             // 1150
if (Meteor.isServer) {                                                                                       // 1151
  var Crypto = Npm.require("crypto");                                                                        // 1152
  var ContentType = Npm.require("content-type");                                                             // 1153
  var Zlib = Npm.require("zlib");                                                                            // 1154
                                                                                                             // 1155
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                                    // 1156
                                                                                                             // 1157
  var computeStagger = function (n) {                                                                        // 1158
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                // 1159
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                 // 1160
    // The sequence looks like:                                                                              // 1161
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                     // 1162
    //                                                                                                       // 1163
    // We use this to determine how we'll stagger periodic events performed by this replica.                 // 1164
    // Notice that this allows us to compute a stagger which is independent of the number of                 // 1165
    // front-end replicas present; we can add more replicas to the end without affecting how the             // 1166
    // earlier ones schedule their events.                                                                   // 1167
    var denom = 1;                                                                                           // 1168
    while (denom <= n) denom <<= 1;                                                                          // 1169
    var num = n * 2 - denom + 1;                                                                             // 1170
    return num / denom;                                                                                      // 1171
  }                                                                                                          // 1172
                                                                                                             // 1173
  var stagger = computeStagger(replicaNumber);                                                               // 1174
                                                                                                             // 1175
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                            // 1176
    // Register a database cleanup function than should run periodically, roughly once every                 // 1177
    // interval of the given length.                                                                         // 1178
    //                                                                                                       // 1179
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be              // 1180
    // scaled appropriately on the assumption that more data is being generated demanding more               // 1181
    // frequent cleanups.                                                                                    // 1182
                                                                                                             // 1183
    check(intervalMs, Number);                                                                               // 1184
    check(callback, Function);                                                                               // 1185
                                                                                                             // 1186
    if (intervalMs < 120000) {                                                                               // 1187
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                               // 1188
                      "are you using the right units?");                                                     // 1189
    }                                                                                                        // 1190
                                                                                                             // 1191
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that              // 1192
    // the schedule is independent of the exact startup time.                                                // 1193
    var first = intervalMs - Date.now() % intervalMs;                                                        // 1194
                                                                                                             // 1195
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the               // 1196
    // same data at the same time.                                                                           // 1197
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                         // 1198
                                                                                                             // 1199
    // If the stagger put us more than an interval away from now, back up.                                   // 1200
    if (first > intervalMs) first -= intervalMs;                                                             // 1201
                                                                                                             // 1202
    Meteor.setTimeout(function () {                                                                          // 1203
      callback();                                                                                            // 1204
      Meteor.setInterval(callback, intervalMs);                                                              // 1205
    }, first);                                                                                               // 1206
  }                                                                                                          // 1207
                                                                                                             // 1208
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                  // 1209
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                          // 1210
                                                                                                             // 1211
  var BufferSmallerThan = function (limit) {                                                                 // 1212
    return Match.Where(function (buf) {                                                                      // 1213
      check(buf, Buffer);                                                                                    // 1214
      return buf.length < limit;                                                                             // 1215
    });                                                                                                      // 1216
  }                                                                                                          // 1217
                                                                                                             // 1218
  var DatabaseId = Match.Where(function (s) {                                                                // 1219
    check(s, String);                                                                                        // 1220
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                     // 1221
  });                                                                                                        // 1222
                                                                                                             // 1223
  addStaticAsset = function (metadata, content) {                                                            // 1224
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it             // 1225
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.          // 1226
                                                                                                             // 1227
    if (typeof content === "string" && !metadata.encoding) {                                                 // 1228
      content = gzipSync(new Buffer(content, "utf8"));                                                       // 1229
      metadata.encoding = "gzip";                                                                            // 1230
    }                                                                                                        // 1231
                                                                                                             // 1232
    check(metadata, {                                                                                        // 1233
      mimeType: String,                                                                                      // 1234
      encoding: Match.Optional("gzip")                                                                       // 1235
    });                                                                                                      // 1236
    check(content, BufferSmallerThan(1 << 20));                                                              // 1237
                                                                                                             // 1238
    // Validate content type.                                                                                // 1239
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                            // 1240
                                                                                                             // 1241
    var hasher = Crypto.createHash("sha256");                                                                // 1242
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                              // 1243
    hasher.update(content);                                                                                  // 1244
    var hash = hasher.digest("base64");                                                                      // 1245
                                                                                                             // 1246
    var existing = StaticAssets.findAndModify({                                                              // 1247
      query: {hash: hash, refcount: {$gte: 1}},                                                              // 1248
      update: {$inc: {refcount: 1}},                                                                         // 1249
      fields: {_id: 1, refcount: 1},                                                                         // 1250
    });                                                                                                      // 1251
    if (existing) {                                                                                          // 1252
      return existing._id;                                                                                   // 1253
    }                                                                                                        // 1254
                                                                                                             // 1255
    return StaticAssets.insert(_.extend({                                                                    // 1256
      hash: hash,                                                                                            // 1257
      content: content,                                                                                      // 1258
      refcount: 1                                                                                            // 1259
    }, metadata));                                                                                           // 1260
  }                                                                                                          // 1261
                                                                                                             // 1262
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                     // 1263
                                                                                                             // 1264
  SandstormDb.prototype.refStaticAsset = function (id) {                                                     // 1265
    // Increment the refcount on an existing static asset.                                                   // 1266
    //                                                                                                       // 1267
    // You must call this BEFORE adding the new reference to the DB, in case of failure between              // 1268
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1269
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1270
                                                                                                             // 1271
    check(id, String);                                                                                       // 1272
                                                                                                             // 1273
    var existing = StaticAssets.findAndModify({                                                              // 1274
      query: {hash: hash},                                                                                   // 1275
      update: {$inc: {refcount: 1}},                                                                         // 1276
      fields: {_id: 1, refcount: 1},                                                                         // 1277
    });                                                                                                      // 1278
    if (!existing) {                                                                                         // 1279
      throw new Error("refStaticAsset() called on asset that doesn't exist");                                // 1280
    }                                                                                                        // 1281
  }                                                                                                          // 1282
                                                                                                             // 1283
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                                   // 1284
    // Decrement refcount on a static asset and delete if it has reached zero.                               // 1285
    //                                                                                                       // 1286
    // You must call this AFTER removing the reference from the DB, in case of failure between               // 1287
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1288
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1289
                                                                                                             // 1290
    check(id, String);                                                                                       // 1291
                                                                                                             // 1292
    var existing = StaticAssets.findAndModify({                                                              // 1293
      query: {_id: id},                                                                                      // 1294
      update: {$inc: {refcount: -1}},                                                                        // 1295
      fields: {_id: 1, refcount: 1},                                                                         // 1296
      new: true,                                                                                             // 1297
    });                                                                                                      // 1298
    if (!existing) {                                                                                         // 1299
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);               // 1300
    } else if (existing.refcount <= 0) {                                                                     // 1301
      StaticAssets.remove({_id: existing._id});                                                              // 1302
    }                                                                                                        // 1303
  }                                                                                                          // 1304
                                                                                                             // 1305
  SandstormDb.prototype.getStaticAsset = function (id) {                                                     // 1306
    // Get a static asset's mimeType, encoding, and raw content.                                             // 1307
                                                                                                             // 1308
    check(id, String);                                                                                       // 1309
                                                                                                             // 1310
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});          // 1311
    if (asset) {                                                                                             // 1312
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy              // 1313
      //   here.                                                                                             // 1314
      asset.content = new Buffer(asset.content);                                                             // 1315
    }                                                                                                        // 1316
    return asset;                                                                                            // 1317
  }                                                                                                          // 1318
                                                                                                             // 1319
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                                // 1320
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: DatabaseId}});                          // 1321
                                                                                                             // 1322
    return AssetUploadTokens.insert({                                                                        // 1323
      purpose: purpose,                                                                                      // 1324
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                               // 1325
    });                                                                                                      // 1326
  }                                                                                                          // 1327
                                                                                                             // 1328
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                 // 1329
    // Indicates that the given asset upload has completed. It will be removed and its purpose               // 1330
    // returned. If no matching upload exists, returns undefined.                                            // 1331
                                                                                                             // 1332
    check(id, String);                                                                                       // 1333
                                                                                                             // 1334
    var upload = AssetUploadTokens.findAndModify({                                                           // 1335
      query: {_id: id},                                                                                      // 1336
      remove: true                                                                                           // 1337
    });                                                                                                      // 1338
                                                                                                             // 1339
    if (upload.expires.valueOf() < Date.now()) {                                                             // 1340
      return undefined;  // already expired                                                                  // 1341
    } else {                                                                                                 // 1342
      return upload.purpose;                                                                                 // 1343
    }                                                                                                        // 1344
  }                                                                                                          // 1345
                                                                                                             // 1346
  function cleanupExpiredAssetUploads() {                                                                    // 1347
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                                  // 1348
  }                                                                                                          // 1349
                                                                                                             // 1350
  // Cleanup tokens every hour.                                                                              // 1351
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                          // 1352
                                                                                                             // 1353
  var packageCache = {};                                                                                     // 1354
  // Package info is immutable. Let's cache to save on mongo queries.                                        // 1355
                                                                                                             // 1356
  SandstormDb.prototype.getPackage = function (packageId) {                                                  // 1357
    // Get the given package record. Since package info is immutable, cache the data in the server           // 1358
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a               // 1359
    // lot.                                                                                                  // 1360
                                                                                                             // 1361
    if (packageId in packageCache) {                                                                         // 1362
      return packageCache[packageId];                                                                        // 1363
    }                                                                                                        // 1364
                                                                                                             // 1365
    var pkg = Packages.findOne(packageId);                                                                   // 1366
    if (pkg && pkg.status === "ready") {                                                                     // 1367
      packageCache[packageId] = pkg;                                                                         // 1368
    }                                                                                                        // 1369
    return pkg;                                                                                              // 1370
  }                                                                                                          // 1371
                                                                                                             // 1372
  SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name,                       // 1373
                                                               versionNumber, marketingVersion) {            // 1374
    var db = this;                                                                                           // 1375
    var actions = db.collections.userActions.find({appId: appId, appVersion: {$lt: versionNumber}},          // 1376
      {fields: {userId: 1}});                                                                                // 1377
    actions.forEach(function (action) {                                                                      // 1378
      var userId = action.userId;                                                                            // 1379
      var updater = {                                                                                        // 1380
        userId: userId,                                                                                      // 1381
        timestamp: new Date(),                                                                               // 1382
        isUnread: true,                                                                                      // 1383
      };                                                                                                     // 1384
                                                                                                             // 1385
      // Set only the appId that we care about. Use mongo's dot notation to specify only a single            // 1386
      // field inside of an object to update                                                                 // 1387
      updater["appUpdates." + appId] = {                                                                     // 1388
        marketingVersion: marketingVersion,                                                                  // 1389
        packageId: packageId,                                                                                // 1390
        name: name,                                                                                          // 1391
        version: versionNumber,                                                                              // 1392
      };                                                                                                     // 1393
      db.collections.notifications.upsert({userId: userId}, { $set: updater});                               // 1394
    });                                                                                                      // 1395
                                                                                                             // 1396
    db.collections.appIndex.update({_id: appId}, {$set: {hasSentNotifications: true}});                      // 1397
                                                                                                             // 1398
    // In the case where we replaced a previous notification and that was the only reference to the          // 1399
    // package, we need to clean it up                                                                       // 1400
    Meteor.call("deleteUnusedPackages", appId);                                                              // 1401
  };                                                                                                         // 1402
                                                                                                             // 1403
  SandstormDb.prototype.sendReferralProgramNotification = sendReferralProgramNotification;                   // 1404
                                                                                                             // 1405
  SandstormDb.prototype.upgradeGrains =  function (appId, version, packageId, backend) {                     // 1406
    check(appId, String);                                                                                    // 1407
    check(version, Match.Integer);                                                                           // 1408
    check(packageId, String);                                                                                // 1409
                                                                                                             // 1410
    var selector = {                                                                                         // 1411
      userId: Meteor.userId(),                                                                               // 1412
      appId: appId,                                                                                          // 1413
      appVersion: { $lte: version },                                                                         // 1414
      packageId: { $ne: packageId }                                                                          // 1415
    };                                                                                                       // 1416
                                                                                                             // 1417
    if (!this.isSimulation) {                                                                                // 1418
      Grains.find(selector).forEach(function (grain) {                                                       // 1419
        backend.shutdownGrain(grain._id, grain.userId);                                                      // 1420
      });                                                                                                    // 1421
    }                                                                                                        // 1422
                                                                                                             // 1423
    Grains.update(selector, { $set: { appVersion: version, packageId: packageId }}, {multi: true});          // 1424
  };                                                                                                         // 1425
                                                                                                             // 1426
  SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {               // 1427
    // Mark package for possible installation.                                                               // 1428
                                                                                                             // 1429
    var fields = {status: "download", progress: 0, url: url, isAutoUpdated: !!isAutoUpdated};                // 1430
                                                                                                             // 1431
    if (retryFailed) {                                                                                       // 1432
      Packages.update({_id: packageId, status: "failed"}, {$set: fields});                                   // 1433
    } else {                                                                                                 // 1434
      try {                                                                                                  // 1435
        fields._id = packageId;                                                                              // 1436
        Packages.insert(fields);                                                                             // 1437
      } catch (err) {                                                                                        // 1438
        console.error("Simultaneous startInstall()s?", err.stack);                                           // 1439
      }                                                                                                      // 1440
    }                                                                                                        // 1441
  };                                                                                                         // 1442
                                                                                                             // 1443
  var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                          // 1444
    check(keyFingerprint, String);                                                                           // 1445
    return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                         // 1446
  });                                                                                                        // 1447
                                                                                                             // 1448
  SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                              // 1449
    // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.           // 1450
                                                                                                             // 1451
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1452
                                                                                                             // 1453
    console.log("fetching keybase", keyFingerprint);                                                         // 1454
                                                                                                             // 1455
    HTTP.get(                                                                                                // 1456
        "https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint +                  // 1457
        "&fields=basics,profile,proofs_summary", {                                                           // 1458
      timeout: 5000                                                                                          // 1459
    }, function (err, keybaseResponse) {                                                                     // 1460
      if (err) {                                                                                             // 1461
        console.log("keybase lookup error:", err.stack);                                                     // 1462
        return;                                                                                              // 1463
      }                                                                                                      // 1464
                                                                                                             // 1465
      if (!keybaseResponse.data) {                                                                           // 1466
        console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                        // 1467
        return;                                                                                              // 1468
      }                                                                                                      // 1469
                                                                                                             // 1470
      var profile = (keybaseResponse.data.them || [])[0];                                                    // 1471
                                                                                                             // 1472
      if (profile) {                                                                                         // 1473
        var record = {                                                                                       // 1474
          displayName: (profile.profile || {}).full_name,                                                    // 1475
          handle: (profile.basics || {}).username,                                                           // 1476
          proofs: (profile.proofs_summary || {}).all || []                                                   // 1477
        };                                                                                                   // 1478
                                                                                                             // 1479
        record.proofs.forEach(function (proof) {                                                             // 1480
          // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that            // 1481
          // this would filter.)                                                                             // 1482
          for (var field in proof) {                                                                         // 1483
            // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly            // 1484
            // because I'm too lazy to check the field names recursively (and Keybase doesn't                // 1485
            // return any objects anyway).                                                                   // 1486
            if (field.match(/[.$]/) || typeof(proof[field]) === "object") {                                  // 1487
              delete proof[field];                                                                           // 1488
            }                                                                                                // 1489
          }                                                                                                  // 1490
                                                                                                             // 1491
          // Indicate not verified.                                                                          // 1492
          // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the            // 1493
          //   Keybase node-based CLI.                                                                       // 1494
          proof.status = "unverified";                                                                       // 1495
        });                                                                                                  // 1496
                                                                                                             // 1497
        KeybaseProfiles.update(keyFingerprint, {$set: record}, {upsert: true});                              // 1498
      } else {                                                                                               // 1499
        // Keybase reports no match, so remove what we know of this user. We don't want to remove            // 1500
        // the item entirely from the cache as this will cause us to repeatedly re-fetch the data            // 1501
        // from Keybase.                                                                                     // 1502
        //                                                                                                   // 1503
        // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,             // 1504
        //   but at present we don't have the ability to verify proofs.                                      // 1505
        KeybaseProfiles.update(keyFingerprint,                                                               // 1506
            {$unset: {displayName: "", handle: "", proofs: ""}}, {upsert: true});                            // 1507
      }                                                                                                      // 1508
    });                                                                                                      // 1509
  };                                                                                                         // 1510
                                                                                                             // 1511
  SandstormDb.prototype.deleteUnusedAccount = function(backend, identityId) {                                // 1512
    // If there is an *unused* account that has `identityId` as a login identity, deletes it.                // 1513
                                                                                                             // 1514
    check(identityId, String);                                                                               // 1515
    var account = Meteor.users.findOne({"loginIdentities.id": identityId});                                  // 1516
    if (account &&                                                                                           // 1517
        account.loginIdentities.length == 1 &&                                                               // 1518
        account.nonloginIdentities.length == 0 &&                                                            // 1519
        !Grains.findOne({userId: account._id}) &&                                                            // 1520
        !ApiTokens.findOne({accountId: account._id}) &&                                                      // 1521
        (!account.plan || account.plan === "free") &&                                                        // 1522
        !account.payments &&                                                                                 // 1523
        !Contacts.findOne({ownerId: account._id})) {                                                         // 1524
      Meteor.users.remove({_id: account._id});                                                               // 1525
      backend.deleteUser(account._id);                                                                       // 1526
    }                                                                                                        // 1527
  }                                                                                                          // 1528
                                                                                                             // 1529
  Meteor.publish("keybaseProfile", function (keyFingerprint) {                                               // 1530
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1531
    var db = this.connection.sandstormDb;                                                                    // 1532
                                                                                                             // 1533
    var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                        // 1534
    if (cursor.count() === 0) {                                                                              // 1535
      // Fire off async update.                                                                              // 1536
      db.updateKeybaseProfileAsync(keyFingerprint);                                                          // 1537
    }                                                                                                        // 1538
                                                                                                             // 1539
    return cursor;                                                                                           // 1540
  });                                                                                                        // 1541
                                                                                                             // 1542
  Meteor.publish("appIndex", function (appId) {                                                              // 1543
    check(appId, String);                                                                                    // 1544
    var db = this.connection.sandstormDb;                                                                    // 1545
    var cursor = db.collections.appIndex.find({_id: appId});                                                 // 1546
    return cursor;                                                                                           // 1547
  });                                                                                                        // 1548
                                                                                                             // 1549
  Meteor.publish("userPackages", function() {                                                                // 1550
    // Users should be able to see packages that are either:                                                 // 1551
    // 1. referenced by one of their userActions                                                             // 1552
    // 2. referenced by one of their grains                                                                  // 1553
    var self = this;                                                                                         // 1554
    var db = this.connection.sandstormDb;                                                                    // 1555
                                                                                                             // 1556
    // Note that package information, once it is in the database, is static. There's no need to              // 1557
    // reactively subscribe to changes to a package since they don't change. It's also unecessary            // 1558
    // to reactively remove a package from the client side when it is removed on the server, or              // 1559
    // when the client stops using it, because the worst case is the client has a small amount               // 1560
    // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                // 1561
    // So, we implement this as a cache: the first time each package ID shows up among the user's            // 1562
    // stuff, we push the package info to the client, and then we never update it.                           // 1563
    //                                                                                                       // 1564
    // Alternatively, we could subscribe to each individual package query, but this would waste              // 1565
    // lots of server-side resources watching for events that will never happen or don't matter.             // 1566
    var hasPackage = {};                                                                                     // 1567
    var refPackage = function (packageId) {                                                                  // 1568
      // Ignore dev apps.                                                                                    // 1569
      if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                    // 1570
                                                                                                             // 1571
      if (!hasPackage[packageId]) {                                                                          // 1572
        hasPackage[packageId] = true;                                                                        // 1573
        var pkg = db.getPackage(packageId);                                                                  // 1574
        if (pkg) {                                                                                           // 1575
          self.added("packages", packageId, pkg);                                                            // 1576
        } else {                                                                                             // 1577
          console.error(                                                                                     // 1578
              "shouldn't happen: missing package referenced by user's stuff:", packageId);                   // 1579
        }                                                                                                    // 1580
      }                                                                                                      // 1581
    };                                                                                                       // 1582
                                                                                                             // 1583
    // package source 1: packages referred to by actions                                                     // 1584
    var actions = db.userActions(this.userId);                                                               // 1585
    var actionsHandle = actions.observe({                                                                    // 1586
      added: function(newAction) {                                                                           // 1587
        refPackage(newAction.packageId);                                                                     // 1588
      },                                                                                                     // 1589
      changed: function(oldAction, newAction) {                                                              // 1590
        refPackage(newAction.packageId);                                                                     // 1591
      }                                                                                                      // 1592
    });                                                                                                      // 1593
                                                                                                             // 1594
    // package source 2: packages referred to by grains directly                                             // 1595
    var grains = db.userGrains(this.userId);                                                                 // 1596
    var grainsHandle = grains.observe({                                                                      // 1597
      added: function(newGrain) {                                                                            // 1598
        refPackage(newGrain.packageId);                                                                      // 1599
      },                                                                                                     // 1600
      changed: function(oldGrain, newGrain) {                                                                // 1601
        refPackage(newGrain.packageId);                                                                      // 1602
      }                                                                                                      // 1603
    });                                                                                                      // 1604
                                                                                                             // 1605
    this.onStop(function () {                                                                                // 1606
      actionsHandle.stop();                                                                                  // 1607
      grainsHandle.stop();                                                                                   // 1608
    });                                                                                                      // 1609
    this.ready();                                                                                            // 1610
  });                                                                                                        // 1611
}                                                                                                            // 1612
                                                                                                             // 1613
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/profile.js                                                                          //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var makeIdenticon;                                                                                           // 17
var httpProtocol;                                                                                            // 18
                                                                                                             // 19
if (Meteor.isServer) {                                                                                       // 20
  SandstormDb.ensureSubscriberHasIdentity = function(publishHandler, identityId) {                           // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber             // 22
    // has a given identity linked. Automatically stops the subscription if the user loses the               // 23
    // identity. Returns a boolean indicating whether the user initially has the identity.                   // 24
                                                                                                             // 25
    var userId = publishHandler.userId;                                                                      // 26
    if (userId === identityId) {                                                                             // 27
      return true;                                                                                           // 28
    } else {                                                                                                 // 29
      var hasIdentityCursor =                                                                                // 30
          Meteor.users.find({$or: [{_id: userId, "loginIdentities.id": identityId},                          // 31
                                   {_id: userId, "nonloginIdentities.id": identityId}]});                    // 32
      if (hasIdentityCursor.count() == 0) {                                                                  // 33
        publishHandler.stop();                                                                               // 34
        return false;                                                                                        // 35
      }                                                                                                      // 36
      var handle = hasIdentityCursor.observe({removed: function () { publishHandler.stop(); }});             // 37
      publishHandler.onStop(function () { handle.stop(); });                                                 // 38
      return true;                                                                                           // 39
    }                                                                                                        // 40
  }                                                                                                          // 41
                                                                                                             // 42
  Meteor.publish("identityProfile", function (identityId) {                                                  // 43
    check(identityId, String);                                                                               // 44
    if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                  // 45
                                                                                                             // 46
    return Meteor.users.find({_id: identityId},                                                              // 47
      {fields: {                                                                                             // 48
        "profile":1,                                                                                         // 49
        "unverifiedEmail":1,                                                                                 // 50
        "expires": 1,                                                                                        // 51
        "createdAt": 1,                                                                                      // 52
                                                                                                             // 53
        "services.dev.name":1,                                                                               // 54
                                                                                                             // 55
        "services.google.id":1,                                                                              // 56
        "services.google.email":1,                                                                           // 57
        "services.google.verified_email":1,                                                                  // 58
        "services.google.name":1,                                                                            // 59
        "services.google.picture":1,                                                                         // 60
        "services.google.gender":1,                                                                          // 61
                                                                                                             // 62
        "services.github.id":1,                                                                              // 63
        "services.github.email":1,                                                                           // 64
        "services.github.emails":1,                                                                          // 65
        "services.github.username":1,                                                                        // 66
                                                                                                             // 67
        "services.email.email":1,                                                                            // 68
      }});                                                                                                   // 69
  }),                                                                                                        // 70
                                                                                                             // 71
  Meteor.publish("accountIdentities", function () {                                                          // 72
    // Maybe this should be folded into the "credentials" subscription?                                      // 73
                                                                                                             // 74
    if (!this.userId) return [];                                                                             // 75
                                                                                                             // 76
    return Meteor.users.find(                                                                                // 77
      {_id: this.userId},                                                                                    // 78
      {fields: {                                                                                             // 79
        "profile": 1,                                                                                        // 80
        "verifiedEmail": 1,                                                                                  // 81
        "loginIdentities": 1,                                                                                // 82
        "nonloginIdentities": 1,                                                                             // 83
        "expires": 1,                                                                                        // 84
        "primaryEmail": 1,                                                                                   // 85
      }});                                                                                                   // 86
  });                                                                                                        // 87
                                                                                                             // 88
  makeIdenticon = function (id) {                                                                            // 89
    // We only make identicons client-side.                                                                  // 90
    return undefined;                                                                                        // 91
  }                                                                                                          // 92
                                                                                                             // 93
  var Url = Npm.require("url");                                                                              // 94
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                   // 95
} else {                                                                                                     // 96
  var identiconCache = {};                                                                                   // 97
                                                                                                             // 98
  makeIdenticon = function (id) {                                                                            // 99
    // Given a cryptographic hash as input, generate an identicon. We always use the user's                  // 100
    // identity ID as the input hash (the same thing passed to apps as the user ID).                         // 101
                                                                                                             // 102
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so               // 103
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out           // 104
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it              // 105
    // anyway to be safe.                                                                                    // 106
    id = id.slice(0, 32);                                                                                    // 107
                                                                                                             // 108
    if (id in identiconCache) {                                                                              // 109
      return identiconCache[id];                                                                             // 110
    }                                                                                                        // 111
                                                                                                             // 112
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.             // 113
    var data = new Identicon(id, 512).toString();                                                            // 114
    var result = "data:image/png;base64," + data;                                                            // 115
    identiconCache[id] = result;                                                                             // 116
    return result;                                                                                           // 117
  }                                                                                                          // 118
                                                                                                             // 119
  httpProtocol = window.location.protocol;                                                                   // 120
}                                                                                                            // 121
                                                                                                             // 122
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                          // 123
                                                                                                             // 124
function staticAssetUrl(id, staticHost) {                                                                    // 125
  if (id) {                                                                                                  // 126
    return staticHost + "/" + id;                                                                            // 127
  } else {                                                                                                   // 128
    return undefined;                                                                                        // 129
  }                                                                                                          // 130
}                                                                                                            // 131
                                                                                                             // 132
function filterHandle(handle) {                                                                              // 133
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                  // 134
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                     // 135
}                                                                                                            // 136
                                                                                                             // 137
function emailToHandle(email) {                                                                              // 138
  // Turn an email address into a handle.                                                                    // 139
                                                                                                             // 140
  if (!email) return undefined;                                                                              // 141
                                                                                                             // 142
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters               // 143
  // on the same account.                                                                                    // 144
  var parts = email.split("@");                                                                              // 145
  var base = filterHandle(parts[0].split("+")[0]);                                                           // 146
                                                                                                             // 147
  var domain = (parts[1]||"").split(".");                                                                    // 148
  if (domain[domain.length - 1] === "name") {                                                                // 149
    // Oh, a .name domain. Let's use                                                                         // 150
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                       // 151
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",                  // 152
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],                 // 153
                         base)) {                                                                            // 154
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                  // 155
    base = filterHandle(domain[0]);                                                                          // 156
  }                                                                                                          // 157
                                                                                                             // 158
  return filterHandle(base);                                                                                 // 159
}                                                                                                            // 160
                                                                                                             // 161
SandstormDb.fillInProfileDefaults = function(user) {                                                         // 162
  var profile = user.profile;                                                                                // 163
  if (profile.service === "github") {                                                                        // 164
    profile.name = profile.name || user.services.github.username || "Name Unknown";                          // 165
    profile.handle = profile.handle || filterHandle(user.services.github.username) ||                        // 166
        filterHandle(profile.name);                                                                          // 167
  } else if (profile.service === "google") {                                                                 // 168
    profile.name = profile.name || user.services.google.name || "Name Unknown";                              // 169
    profile.handle = profile.handle || emailToHandle(user.services.google.email) ||                          // 170
        filterHandle(profile.name);                                                                          // 171
    profile.pronoun = profile.pronoun || GENDERS[user.services.google.gender] || "neutral";                  // 172
  } else if (profile.service === "email") {                                                                  // 173
    var email = user.services.email.email                                                                    // 174
    profile.name = profile.name || email.split("@")[0];                                                      // 175
    profile.handle = profile.handle || emailToHandle(email);                                                 // 176
  } else if (profile.service === "dev") {                                                                    // 177
    var lowerCaseName = user.services.dev.name.split(" ")[0].toLowerCase();                                  // 178
    profile.name = profile.name || user.services.dev.name;                                                   // 179
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                          // 180
    profile.pronoun = profile.pronoun ||                                                                     // 181
        (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" :                                   // 182
         _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");                                   // 183
  } else if (profile.service === "demo") {                                                                   // 184
    profile.name = profile.name || "Demo User";                                                              // 185
    profile.handle = profile.handle || "demo";                                                               // 186
  } else {                                                                                                   // 187
    throw new Error("unrecognized identity service: ", profile.service);                                     // 188
  }                                                                                                          // 189
                                                                                                             // 190
  profile.pronoun = profile.pronoun || "neutral";                                                            // 191
}                                                                                                            // 192
                                                                                                             // 193
SandstormDb.fillInIntrinsicName = function(user) {                                                           // 194
  var profile = user.profile;                                                                                // 195
  if (profile.service === "github") {                                                                        // 196
    profile.intrinsicName = user.services.github.username;                                                   // 197
  } else if (profile.service === "google") {                                                                 // 198
    profile.intrinsicName = user.services.google.name;                                                       // 199
    user.privateIntrinsicName = user.services.google.email;                                                  // 200
  } else if (profile.service === "email") {                                                                  // 201
    profile.intrinsicName = user.services.email.email                                                        // 202
  } else if (profile.service === "dev") {                                                                    // 203
    profile.intrinsicName = user.services.dev.name;                                                          // 204
  } else if (profile.service === "demo") {                                                                   // 205
    profile.intrinsicName = "demo on " + user.createdAt.toISOString().substring(0,10);                       // 206
  } else {                                                                                                   // 207
    throw new Error("unrecognized identity service: ", profile.service);                                     // 208
  }                                                                                                          // 209
}                                                                                                            // 210
                                                                                                             // 211
SandstormDb.getVerifiedEmails = function(identity) {                                                         // 212
  if (identity.services.google && identity.services.google.email &&                                          // 213
      identity.services.google.verified_email) {                                                             // 214
    return [identity.services.google.email];                                                                 // 215
  } else if (identity.services.email) {                                                                      // 216
    return [identity.services.email.email];                                                                  // 217
  } else if (identity.services.github && identity.services.github.emails) {                                  // 218
    return _.pluck(_.filter(identity.services.github.emails,                                                 // 219
                            function(email) { return email.verified; }),                                     // 220
                   "email");                                                                                 // 221
  }                                                                                                          // 222
  return [];                                                                                                 // 223
}                                                                                                            // 224
                                                                                                             // 225
SandstormDb.fillInPictureUrl = function(user) {                                                              // 226
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                         // 227
  user.profile.pictureUrl = staticAssetUrl(user.profile.picture, staticHost) ||                              // 228
    makeIdenticon(user._id);                                                                                 // 229
}                                                                                                            // 230
                                                                                                             // 231
SandstormDb.getUserIdentityIds = function (user) {                                                           // 232
  // Given an account user object, returns an array containing the ID of each identity linked to the         // 233
  // account. Always returns the most recently added login identity first.                                   // 234
  if (user && user.loginIdentities) {                                                                        // 235
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                    // 236
  } else {                                                                                                   // 237
    return [];                                                                                               // 238
  }                                                                                                          // 239
};                                                                                                           // 240
                                                                                                             // 241
SandstormDb.getUserEmails = function (user) {                                                                // 242
  // Given a user object, returns an array containing all email addresses associated with that user.         // 243
  // Each entry in the array is an object of the form:                                                       // 244
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                          // 245
  //                                                                                                         // 246
  // At most one entry in the result has `primary = true`.                                                   // 247
  //                                                                                                         // 248
  // TODO(cleanup): This actually does need to query the database to fetch profile information               // 249
  //   for linked identities, so it probably makes more sense for it to be a non-static method               // 250
  //   on SandstormDb.                                                                                       // 251
                                                                                                             // 252
  var identityIds = SandstormDb.getUserIdentityIds(user);                                                    // 253
  verifiedEmails = {};                                                                                       // 254
  unverifiedEmails = {};                                                                                     // 255
                                                                                                             // 256
  identityIds.forEach(function (id) {                                                                        // 257
    var identity = Meteor.users.findOne({_id: id});                                                          // 258
    if (identity && identity.services) {                                                                     // 259
      SandstormDb.getVerifiedEmails(identity).forEach(function(verifiedEmail) {                              // 260
        if (verifiedEmail) {                                                                                 // 261
          verifiedEmails[verifiedEmail] = true;                                                              // 262
        }                                                                                                    // 263
      });                                                                                                    // 264
    }                                                                                                        // 265
    if (identity && identity.unverifiedEmail) {                                                              // 266
      unverifiedEmails[identity.unverifiedEmail] = true;                                                     // 267
    }                                                                                                        // 268
  });                                                                                                        // 269
                                                                                                             // 270
  var result = [];                                                                                           // 271
  _.keys(verifiedEmails).map(function (email) {                                                              // 272
    result.push({email: email,                                                                               // 273
                 verified: true,                                                                             // 274
                 primary: email === user.primaryEmail});                                                     // 275
  });                                                                                                        // 276
  _.keys(unverifiedEmails).map(function (email) {                                                            // 277
    if (!(email in verifiedEmails)) { result.push({email: email, verified: false}) }                         // 278
  });                                                                                                        // 279
                                                                                                             // 280
  // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                     // 281
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                         // 282
    result[0].primary = true;                                                                                // 283
  }                                                                                                          // 284
  return result;                                                                                             // 285
}                                                                                                            // 286
                                                                                                             // 287
SandstormDb.prototype.addContact = function addContact (ownerId, identityId) {                               // 288
  check(ownerId, String);                                                                                    // 289
  check(identityId, String);                                                                                 // 290
  var db = this;                                                                                             // 291
  var profile = db.getIdentity(identityId).profile;                                                          // 292
  db.collections.contacts.upsert({ownerId: ownerId, identityId: identityId}, {                               // 293
    ownerId: ownerId,                                                                                        // 294
    petname: profile && profile.name,                                                                        // 295
    created: new Date(),                                                                                     // 296
    identityId: identityId,                                                                                  // 297
  });                                                                                                        // 298
}                                                                                                            // 299
                                                                                                             // 300
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/user.js                                                                             //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var Crypto = Npm.require("crypto");                                                                          // 17
var Future = Npm.require("fibers/future");                                                                   // 18
                                                                                                             // 19
userPictureUrl = function (user) {                                                                           // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                            // 21
    // Try to determine user's avatar URL from login service.                                                // 22
                                                                                                             // 23
    var google = user.services.google;                                                                       // 24
    if (google && google.picture) {                                                                          // 25
      return google.picture;                                                                                 // 26
    }                                                                                                        // 27
                                                                                                             // 28
    var github = user.services.github;                                                                       // 29
    if (github && github.id) {                                                                               // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                         // 31
    }                                                                                                        // 32
                                                                                                             // 33
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be            // 34
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                // 35
    // Github are different because they are actually the identity providers, so they already know           // 36
    // the user logged in.                                                                                   // 37
  }                                                                                                          // 38
}                                                                                                            // 39
                                                                                                             // 40
fetchPicture = function (url) {                                                                              // 41
  try {                                                                                                      // 42
    var result = HTTP.get(url, {                                                                             // 43
      npmRequestOptions: { encoding: null },                                                                 // 44
      timeout: 5000                                                                                          // 45
    });                                                                                                      // 46
                                                                                                             // 47
    var metadata = {};                                                                                       // 48
                                                                                                             // 49
    metadata.mimeType = result.headers["content-type"];                                                      // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                              // 51
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                             // 52
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                        // 53
    }                                                                                                        // 54
    var enc = result.headers["content-encoding"];                                                            // 55
    if (enc && enc !== "identity") {                                                                         // 56
      metadata.encoding = enc;                                                                               // 57
    }                                                                                                        // 58
                                                                                                             // 59
    return addStaticAsset(metadata, result.content);                                                         // 60
  } catch (err) {                                                                                            // 61
    console.error("failed to fetch user profile picture:", url, err.stack);                                  // 62
  }                                                                                                          // 63
}                                                                                                            // 64
                                                                                                             // 65
var ValidHandle = Match.Where(function (handle) {                                                            // 66
  check(handle, String);                                                                                     // 67
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                               // 68
});                                                                                                          // 69
                                                                                                             // 70
Accounts.onCreateUser(function (options, user) {                                                             // 71
  if (user.loginIdentities) {                                                                                // 72
    // it's an account                                                                                       // 73
    check(user, {_id: String,                                                                                // 74
                 createdAt: Date,                                                                            // 75
                 isAdmin: Match.Optional(Boolean),                                                           // 76
                 hasCompletedSignup: Match.Optional(Boolean),                                                // 77
                 signupKey: Match.Optional(String),                                                          // 78
                 signupNote: Match.Optional(String),                                                         // 79
                 signupEmail: Match.Optional(String),                                                        // 80
                 expires: Match.Optional(Date),                                                              // 81
                 appDemoId: Match.Optional(String),                                                          // 82
                 loginIdentities: [{id: String}],                                                            // 83
                 nonloginIdentities: [{id: String}]});                                                       // 84
                                                                                                             // 85
    if (Meteor.settings.public.quotaEnabled) {                                                               // 86
      user.experiments = user.experiments || {};                                                             // 87
      user.experiments = {                                                                                   // 88
        firstTimeBillingPrompt: Math.random() < 0.5 ? "control" : "test"                                     // 89
      };                                                                                                     // 90
      if (!("expires" in user)) {                                                                            // 91
        sendReferralProgramNotification(user._id);                                                           // 92
      }                                                                                                      // 93
    }                                                                                                        // 94
                                                                                                             // 95
    return user;                                                                                             // 96
  }                                                                                                          // 97
                                                                                                             // 98
  // Check profile.                                                                                          // 99
  if (options.profile) {                                                                                     // 100
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                     // 101
    check(options.profile, Match.ObjectIncluding({                                                           // 102
      name: Match.OneOf(null, Match.Optional(String)),                                                       // 103
      handle: Match.Optional(ValidHandle),                                                                   // 104
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                            // 105
    }));                                                                                                     // 106
  }                                                                                                          // 107
                                                                                                             // 108
  check(options.unverifiedEmail, Match.Optional(String));                                                    // 109
                                                                                                             // 110
  if (options.unverifiedEmail) {                                                                             // 111
    user.unverifiedEmail = options.unverifiedEmail;                                                          // 112
  }                                                                                                          // 113
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronouns");                                // 114
                                                                                                             // 115
  // Try downloading avatar.                                                                                 // 116
  var url = userPictureUrl(user);                                                                            // 117
  if (url) {                                                                                                 // 118
    var assetId = fetchPicture(url);                                                                         // 119
    if (assetId) {                                                                                           // 120
      user.profile.picture = assetId;                                                                        // 121
    }                                                                                                        // 122
  }                                                                                                          // 123
                                                                                                             // 124
  var serviceUserId;                                                                                         // 125
  if (user.services && user.services.dev) {                                                                  // 126
    check(user.services.dev, {name: String, isAdmin: Boolean, hasCompletedSignup: Boolean});                 // 127
    serviceUserId = user.services.dev.name;                                                                  // 128
    user.profile.service = "dev";                                                                            // 129
  } else if ("expires" in user) {                                                                            // 130
    serviceUserId = user._id;                                                                                // 131
    user.profile.service = "demo";                                                                           // 132
  } else if (user.services && user.services.email) {                                                         // 133
    check(user.services.email,                                                                               // 134
          {email: String,                                                                                    // 135
           tokens: [{digest: String, algorithm: String, createdAt: Date}]});                                 // 136
    serviceUserId = user.services.email.email;                                                               // 137
    user.profile.service = "email";                                                                          // 138
  } else if (user.services && "google" in user.services) {                                                   // 139
    serviceUserId = user.services.google.id;                                                                 // 140
    user.profile.service = "google";                                                                         // 141
  } else if (user.services && "github" in user.services) {                                                   // 142
    serviceUserId = user.services.github.id;                                                                 // 143
    user.profile.service = "github";                                                                         // 144
  } else {                                                                                                   // 145
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " +                      // 146
                           JSON.stringify(user));                                                            // 147
  }                                                                                                          // 148
  user._id = Crypto.createHash("sha256")                                                                     // 149
    .update(user.profile.service + ":" + serviceUserId).digest("hex");                                       // 150
                                                                                                             // 151
  return user;                                                                                               // 152
});                                                                                                          // 153
                                                                                                             // 154
// TODO delete this obsolete index.                                                                          // 155
Meteor.users._ensureIndex("identities.id", {unique: 1, sparse: 1});                                          // 156
                                                                                                             // 157
Meteor.users._ensureIndex("loginIdentities.id", {unique: 1, sparse: 1});                                     // 158
Meteor.users._ensureIndex("nonloginIdentities.id", {sparse: 1});                                             // 159
                                                                                                             // 160
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/migrations.js                                                                       //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                          // 1
//                                                                                                           // 2
// Because migrations can experience partial failure and likely have                                         // 3
// side-effects, we should be careful to make sure all migrations are                                        // 4
// idempotent and safe to accidentally run multiple times.                                                   // 5
                                                                                                             // 6
var Future = Npm.require("fibers/future");                                                                   // 7
var Url = Npm.require("url");                                                                                // 8
                                                                                                             // 9
var updateLoginStyleToRedirect = function() {                                                                // 10
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 11
  ["google", "github"].forEach(function(serviceName) {                                                       // 12
    var config = configurations.findOne({service: serviceName});                                             // 13
    if (config && config.loginStyle !== "redirect") {                                                        // 14
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                       // 15
    }                                                                                                        // 16
  });                                                                                                        // 17
};                                                                                                           // 18
                                                                                                             // 19
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                                 // 20
  // In the before time, Google and Github login were enabled by default.                                    // 21
  //                                                                                                         // 22
  // This actually didn't make much sense, required the first user to configure                              // 23
  // OAuth, and had some trust-the-first-user properties that weren't totally                                // 24
  // secure.                                                                                                 // 25
  //                                                                                                         // 26
  // Now, we have admin-token, so we wish to disable all logins by default                                   // 27
  // (since they need to be configured anyway) but since users may have never                                // 28
  // explicitly told Sandstorm that Google or Github login should be enabled,                                // 29
  // we can't just use the value in the Settings collection, since it might                                  // 30
  // never have been set.                                                                                    // 31
  //                                                                                                         // 32
  // Thus, if the service is configured but absent from Settings, we should                                  // 33
  // explicitly enable it in Settings, and then the rest of the logic can just                               // 34
  // depend on what value is in Settings and default to false without breaking                               // 35
  // user installations.                                                                                     // 36
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 37
  ["google", "github"].forEach(function(serviceName) {                                                       // 38
    var config = configurations.findOne({service: serviceName});                                             // 39
    var serviceConfig = Settings.findOne({_id: serviceName});                                                // 40
    if (config && !serviceConfig) {                                                                          // 41
      // Only explicitly enable the login service if:                                                        // 42
      // 1) the service is already configured                                                                // 43
      // 2) there is no sandstorm configuration already present (the user was                                // 44
      //    using the previous default behavior).                                                            // 45
      Settings.insert({_id: serviceName, value: true});                                                      // 46
    }                                                                                                        // 47
  });                                                                                                        // 48
};                                                                                                           // 49
                                                                                                             // 50
var denormalizeInviteInfo = function() {                                                                     // 51
  // When a user is invited via a signup token, the `signupKey` field of their user table entry              // 52
  // has always been populated to indicate the key they used. This points into the SignupKeys table          // 53
  // which has more information about the key, namely a freeform note entered by the admin when              // 54
  // they created the key. In the case that the email invite form was used, the note has the form            // 55
  // "E-mail invite to <address>".                                                                           // 56
  //                                                                                                         // 57
  // Later, we decided it was useful to indicate in the users table visible to the admin                     // 58
  // information about the invite terms. Namely, for email invites we want to show the address               // 59
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                // 60
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing              // 61
  // from the SignupKeys table.                                                                              // 62
                                                                                                             // 63
  Meteor.users.find().forEach(function (user) {                                                              // 64
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {              // 65
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                   // 66
      if (signupInfo && signupInfo.note) {                                                                   // 67
        var newFields = { signupNote: signupInfo.note };                                                     // 68
                                                                                                             // 69
        var prefix = "E-mail invite to ";                                                                    // 70
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                     // 71
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                      // 72
        }                                                                                                    // 73
                                                                                                             // 74
        Meteor.users.update(user._id, {$set: newFields});                                                    // 75
      }                                                                                                      // 76
    }                                                                                                        // 77
  });                                                                                                        // 78
}                                                                                                            // 79
                                                                                                             // 80
function mergeRoleAssignmentsIntoApiTokens() {                                                               // 81
  RoleAssignments.find().forEach(function(roleAssignment) {                                                  // 82
    ApiTokens.insert({                                                                                       // 83
      grainId: roleAssignment.grainId,                                                                       // 84
      userId: roleAssignment.sharer,                                                                         // 85
      roleAssignment: roleAssignment.roleAssignment,                                                         // 86
      petname: roleAssignment.petname,                                                                       // 87
      created: roleAssignment.created,                                                                       // 88
      owner: {user: {userId: roleAssignment.recipient,                                                       // 89
                     title: roleAssignment.title}},                                                          // 90
    });                                                                                                      // 91
  });                                                                                                        // 92
}                                                                                                            // 93
                                                                                                             // 94
function fixOasisStorageUsageStats() {}                                                                      // 95
// This migration only pertained to Oasis and it was successfully applied there. Since it referred           // 96
// to some global variables that we later wanted to remove and/or rename, we've since replaced it            // 97
// with a no-op.                                                                                             // 98
                                                                                                             // 99
function fetchProfilePictures() {                                                                            // 100
  Meteor.users.find({}).forEach(function (user) {                                                            // 101
    var url = userPictureUrl(user);                                                                          // 102
    if (url) {                                                                                               // 103
      console.log("Fetching user picture:", url);                                                            // 104
      var assetId = fetchPicture(url);                                                                       // 105
      if (assetId) {                                                                                         // 106
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                                 // 107
      }                                                                                                      // 108
    }                                                                                                        // 109
  });                                                                                                        // 110
}                                                                                                            // 111
                                                                                                             // 112
function assignPlans() {                                                                                     // 113
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                            // 114
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                     // 115
  }                                                                                                          // 116
}                                                                                                            // 117
                                                                                                             // 118
function removeKeyrings() {                                                                                  // 119
  // These blobs full of public keys were not intended to find their way into mongo and while                // 120
  // harmless they slow things down because they're huge. Remove them.                                       // 121
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                         // 122
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                        // 123
      {multi: true});                                                                                        // 124
}                                                                                                            // 125
                                                                                                             // 126
function useLocalizedTextInUserActions() {                                                                   // 127
  function toLocalizedText(newObj, oldObj, field) {                                                          // 128
    if (field in oldObj) {                                                                                   // 129
      if (typeof oldObj[field] === "string") {                                                               // 130
        newObj[field] = {defaultText: oldObj[field]};                                                        // 131
      } else {                                                                                               // 132
        newObj[field] = oldObj[field];                                                                       // 133
      }                                                                                                      // 134
    }                                                                                                        // 135
  }                                                                                                          // 136
  UserActions.find({}).forEach(function (userAction) {                                                       // 137
    var fields = {};                                                                                         // 138
    toLocalizedText(fields, userAction, "appTitle");                                                         // 139
    toLocalizedText(fields, userAction, "title");                                                            // 140
    toLocalizedText(fields, userAction, "nounPhrase");                                                       // 141
    UserActions.update(userAction._id, {$set: fields});                                                      // 142
  });                                                                                                        // 143
}                                                                                                            // 144
                                                                                                             // 145
function verifyAllPgpSignatures() {                                                                          // 146
  Packages.find({}).forEach(function (pkg) {                                                                 // 147
    try {                                                                                                    // 148
      console.log("checking PGP signature for package:", pkg._id);                                           // 149
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                    // 150
      if (info.authorPgpKeyFingerprint) {                                                                    // 151
        console.log("  " + info.authorPgpKeyFingerprint);                                                    // 152
        Packages.update(pkg._id,                                                                             // 153
            {$set: {authorPgpKeyFingerprint: info.authorPgpKeyFingerprint}});                                // 154
      } else {                                                                                               // 155
        console.log("  no signature");                                                                       // 156
      }                                                                                                      // 157
    } catch (err) {                                                                                          // 158
      console.error(err.stack);                                                                              // 159
    }                                                                                                        // 160
  });                                                                                                        // 161
}                                                                                                            // 162
                                                                                                             // 163
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                        // 164
  var Crypto = Npm.require("crypto");                                                                        // 165
  Meteor.users.find().forEach(function (user) {                                                              // 166
    var identity = {};                                                                                       // 167
    var serviceUserId;                                                                                       // 168
    if ("devName" in user) {                                                                                 // 169
      identity.service = "dev";                                                                              // 170
      serviceUserId = user.devName;                                                                          // 171
    } else if ("expires" in user) {                                                                          // 172
      identity.service = "demo";                                                                             // 173
      serviceUserId = user._id;                                                                              // 174
    } else if (user.services && "google" in user.services) {                                                 // 175
      identity.service = "google";                                                                           // 176
      if (user.services.google.email && user.services.google.verified_email) {                               // 177
        identity.verifiedEmail = user.services.google.email;                                                 // 178
      }                                                                                                      // 179
      serviceUserId = user.services.google.id;                                                               // 180
    } else if (user.services && "github" in user.services) {                                                 // 181
      identity.service = "github";                                                                           // 182
      identity.unverifiedEmail = user.services.github.email;                                                 // 183
      serviceUserId = user.services.github.id;                                                               // 184
    } else if (user.services && "emailToken" in user.services) {                                             // 185
      identity.service = "emailToken";                                                                       // 186
      identity.verifiedEmail = user.services.emailToken.email;                                               // 187
      serviceUserId = user.services.emailToken.email;                                                        // 188
    }                                                                                                        // 189
                                                                                                             // 190
    identity.id = Crypto.createHash("sha256")                                                                // 191
        .update(identity.service + ":" + serviceUserId).digest("hex");                                       // 192
                                                                                                             // 193
    if (user.profile) {                                                                                      // 194
      if (user.profile.name) {                                                                               // 195
        identity.name = user.profile.name;                                                                   // 196
      }                                                                                                      // 197
      if (user.profile.handle) {                                                                             // 198
        identity.handle = user.profile.handle;                                                               // 199
      }                                                                                                      // 200
      if (user.profile.picture) {                                                                            // 201
        identity.picture = user.profile.picture;                                                             // 202
      }                                                                                                      // 203
      if (user.profile.pronoun) {                                                                            // 204
        identity.pronoun = user.profile.pronoun;                                                             // 205
      }                                                                                                      // 206
      if (user.profile.email) {                                                                              // 207
        identity.unverifiedEmail = user.profile.email;                                                       // 208
      }                                                                                                      // 209
    }                                                                                                        // 210
    identity.main = true;                                                                                    // 211
                                                                                                             // 212
    Meteor.users.update(user._id, {$set: {identities: [identity]}});                                         // 213
                                                                                                             // 214
    Grains.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                     // 215
    Sessions.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                   // 216
    ApiTokens.update({userId: user._id},                                                                     // 217
                     {$set: {identityId: identity.id}},                                                      // 218
                     {multi: true});                                                                         // 219
    ApiTokens.update({"owner.user.userId": user._id},                                                        // 220
                     {$set: {"owner.user.identityId": identity.id}},                                         // 221
                     {multi: true});                                                                         // 222
    ApiTokens.update({"owner.grain.introducerUser": user._id},                                               // 223
                     {$set: {"owner.grain.introducerIdentity": identity.id}},                                // 224
                     {multi: true});                                                                         // 225
                                                                                                             // 226
    while (ApiTokens.update({"requirements.permissionsHeld.userId": user._id},                               // 227
                            {$set: {"requirements.$.permissionsHeld.identityId": identity.id},               // 228
                             $unset: {"requirements.$.permissionsHeld.userId": 1}},                          // 229
                            {multi: true}) > 0);                                                             // 230
    // The `$` operatorer modifies the first element in the array that matches the query. Since              // 231
    // there may be many matches, we need to repeat until no documents are modified.                         // 232
                                                                                                             // 233
  });                                                                                                        // 234
                                                                                                             // 235
  ApiTokens.remove({userInfo: {$exists: true}});                                                             // 236
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                    // 237
  // that this field could show up in the database was in this deprecated, no-longer-functional              // 238
  // form of API token.                                                                                      // 239
}                                                                                                            // 240
                                                                                                             // 241
function appUpdateSettings() {                                                                               // 242
  Settings.insert({_id: "appMarketUrl", value: "https://apps.sandstorm.io"});                                // 243
  Settings.insert({_id: "appIndexUrl", value: "https://app-index.sandstorm.io"});                            // 244
  Settings.insert({_id: "appUpdatesEnabled", value: true});                                                  // 245
}                                                                                                            // 246
                                                                                                             // 247
function moveDevAndEmailLoginDataIntoIdentities() {                                                          // 248
  Meteor.users.find().forEach(function (user) {                                                              // 249
    if (user.identities.length != 1) {                                                                       // 250
      throw new Error("User does not have exactly one identity: ", user);                                    // 251
    }                                                                                                        // 252
    var identity = user.identities[0];                                                                       // 253
    if (Match.test(identity.service, Object)) { return; } // Already migrated.                               // 254
                                                                                                             // 255
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");         // 256
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                          // 257
                                                                                                             // 258
    var serviceObject = {};                                                                                  // 259
    var fieldsToUnset = {}                                                                                   // 260
                                                                                                             // 261
    if (identity.service === "dev") {                                                                        // 262
      serviceObject.name = user.devName;                                                                     // 263
      fieldsToUnset.devName = 1;                                                                             // 264
    } else if (identity.service === "emailToken") {                                                          // 265
      serviceObject.tokens = user.services.emailToken.tokens;                                                // 266
      serviceObject.email = user.services.emailToken.email;                                                  // 267
      fieldsToUnset["services.emailToken"] = 1;                                                              // 268
    }                                                                                                        // 269
                                                                                                             // 270
    newIdentity.service = {};                                                                                // 271
    newIdentity.service[identity.service] = serviceObject;                                                   // 272
                                                                                                             // 273
    var modifier = {$set: {identities: [newIdentity]}};                                                      // 274
    if (Object.keys(fieldsToUnset).length > 0) {                                                             // 275
      modifier["$unset"] = fieldsToUnset;                                                                    // 276
    }                                                                                                        // 277
                                                                                                             // 278
    Meteor.users.update({_id: user._id}, modifier);                                                          // 279
  });                                                                                                        // 280
}                                                                                                            // 281
                                                                                                             // 282
function repairEmailIdentityIds() {                                                                          // 283
  var Crypto = Npm.require("crypto");                                                                        // 284
  Meteor.users.find({"identities.service.emailToken": {$exists: 1}}).forEach(function (user) {               // 285
    if (user.identities.length != 1) {                                                                       // 286
      throw new Error("User does not have exactly one identity: ", user);                                    // 287
    }                                                                                                        // 288
    var identity = user.identities[0];                                                                       // 289
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail",                 // 290
                             "profile");                                                                     // 291
    newIdentity.service = {email: identity.service.emailToken};                                              // 292
    newIdentity.id = Crypto.createHash("sha256")                                                             // 293
      .update("email:" + identity.service.emailToken.email).digest("hex");                                   // 294
                                                                                                             // 295
    Grains.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});           // 296
    Sessions.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});         // 297
    ApiTokens.update({identityId: identity.id},                                                              // 298
                     {$set: {identityId: newIdentity.id}},                                                   // 299
                     {multi: true});                                                                         // 300
    ApiTokens.update({"owner.user.identityId": identity.id},                                                 // 301
                     {$set: {"owner.user.identityId": newIdentity.id}},                                      // 302
                     {multi: true});                                                                         // 303
    ApiTokens.update({"owner.grain.introducerIdentity": identity.id},                                        // 304
                     {$set: {"owner.grain.introducerIdentity": newIdentity.id}},                             // 305
                     {multi: true});                                                                         // 306
                                                                                                             // 307
    while (ApiTokens.update({"requirements.permissionsHeld.identityId": identity.id},                        // 308
                            {$set: {"requirements.$.permissionsHeld.identityId": newIdentity.id}},           // 309
                            {multi: true}) > 0);                                                             // 310
                                                                                                             // 311
    Meteor.users.update({_id: user._id}, {$set: {identities: [newIdentity]}});                               // 312
  });                                                                                                        // 313
}                                                                                                            // 314
                                                                                                             // 315
function splitAccountUsersAndIdentityUsers() {                                                               // 316
  Meteor.users.find({"identities": {$exists: true}}).forEach(function (user) {                               // 317
    if (user.identities.length != 1) {                                                                       // 318
      throw new Error("User does not have exactly one identity: ", user);                                    // 319
    }                                                                                                        // 320
    var identity = user.identities[0];                                                                       // 321
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires")                                    // 322
    identityUser._id = identity.id;                                                                          // 323
    identityUser.profile = identity.profile;                                                                 // 324
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                             // 325
    identityUser.profile.service = Object.keys(identity.service)[0];                                         // 326
                                                                                                             // 327
    // Updating this user needs to be a two step process because the `services` field typically              // 328
    // contains subfields that are constrained to be unique by Mongo indices.                                // 329
    identityUser.stagedServices = _.omit(user.services, "resume");                                           // 330
    if (identity.service.dev) {                                                                              // 331
      identityUser.stagedServices.dev = identity.service.dev;                                                // 332
    } else if (identity.service.email) {                                                                     // 333
      identityUser.stagedServices.email = identity.service.email;                                            // 334
    }                                                                                                        // 335
                                                                                                             // 336
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires",                              // 337
                             "isAdmin", "signupKey", "signupNote", "signupEmail",                            // 338
                             "plan", "storageUsage", "isAppDemoUser", "appDemoId",                           // 339
                             "payments", "dailySentMailCount", "hasCompletedSignup");                        // 340
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                  // 341
    accountUser.nonloginIdentities = [];                                                                     // 342
    if (user.services && user.services.resume) {                                                             // 343
      accountUser.services = {resume: user.services.resume};                                                 // 344
    }                                                                                                        // 345
                                                                                                             // 346
    accountUser.stashedOldUser = user;                                                                       // 347
                                                                                                             // 348
    ApiTokens.update({identityId: identityUser._id}, {$set: {accountId: user._id}},                          // 349
                     {multi: true});                                                                         // 350
                                                                                                             // 351
    Meteor.users.upsert({_id: identityUser._id}, identityUser);                                              // 352
    Meteor.users.update({_id: user._id}, accountUser);                                                       // 353
  });                                                                                                        // 354
                                                                                                             // 355
  Meteor.users.find({"stagedServices": {$exists: true}}).forEach(function (identity) {                       // 356
    Meteor.users.update({_id: identity._id}, {$unset: {stagedServices: 1},                                   // 357
                                              $set: {services: identity.stagedServices}});                   // 358
  });                                                                                                        // 359
}                                                                                                            // 360
                                                                                                             // 361
function populateContactsFromApiTokens() {                                                                   // 362
  ApiTokens.find({"owner.user.identityId": {$exists: 1},                                                     // 363
                  accountId: {$exists: 1}}).forEach(function(token) {                                        // 364
    var identityId = token.owner.user.identityId;                                                            // 365
    var identity = SandstormDb.prototype.getIdentity(identityId);                                            // 366
    if (identity) {                                                                                          // 367
      var profile = identity.profile;                                                                        // 368
      Contacts.upsert({ownerId: token.accountId, identityId: identityId}, {                                  // 369
        ownerId: token.accountId,                                                                            // 370
        petname: profile && profile.name,                                                                    // 371
        created: new Date(),                                                                                 // 372
        identityId: identityId,                                                                              // 373
        profile: profile,                                                                                    // 374
      });                                                                                                    // 375
    }                                                                                                        // 376
  });                                                                                                        // 377
}                                                                                                            // 378
                                                                                                             // 379
function cleanUpApiTokens() {                                                                                // 380
  // The `splitUserIdsIntoAccountIdsAndIdentityIds()` migration only added `identityId` in cases             // 381
  // where the user still existed in the database.                                                           // 382
  ApiTokens.remove({userId: {$exists: true}, identityId: {$exists: false}});                                 // 383
  ApiTokens.remove({"owner.user.userId": {$exists: true},                                                    // 384
                    "owner.user.identityId": {$exists: false}});                                             // 385
                                                                                                             // 386
  // For a while we were accidentally setting `appIcon` instead of `icon`.                                   // 387
  ApiTokens.find({"owner.user.denormalizedGrainMetadata.appIcon": {$exists: true}}).forEach(                 // 388
      function (apiToken) {                                                                                  // 389
    var icon = apiToken.owner.user.denormalizedGrainMetadata.appIcon;                                        // 390
    ApiTokens.update({_id: apiToken._id},                                                                    // 391
                     {$set: {"owner.user.denormalizedGrainMetadata.icon": icon},                             // 392
                      $unset: {"owner.user.denormalizedGrainMetadata.appIcon": true}});                      // 393
  });                                                                                                        // 394
                                                                                                             // 395
  // For a while the `identityId` field of child UiView tokens was not getting set.                          // 396
  function repairChain(parentToken) {                                                                        // 397
    ApiTokens.find({parentToken: parentToken._id, grainId: {$exists: true},                                  // 398
                    identityId: {$exists: false}}).forEach(function (childToken) {                           // 399
      ApiTokens.update({_id: childToken._id}, {$set: {identityId: parentToken.identityId}});                 // 400
      repairChain(childToken);                                                                               // 401
    });                                                                                                      // 402
  }                                                                                                          // 403
  ApiTokens.find({grainId: {$exists: true}, identityId: {$exists: true},                                     // 404
                  parentToken: {$exists: false}}).forEach(repairChain);                                      // 405
}                                                                                                            // 406
                                                                                                             // 407
function initServerTitleAndReturnAddress() {                                                                 // 408
  var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                   // 409
  Settings.insert({_id: "serverTitle", value: hostname});                                                    // 410
  Settings.insert({_id: "returnAddress", value: "no-reply@" + hostname});                                    // 411
}                                                                                                            // 412
                                                                                                             // 413
function sendReferralNotifications() {                                                                       // 414
  if (Meteor.settings.public.quotaEnabled) {                                                                 // 415
    Meteor.users.find({                                                                                      // 416
      loginIdentities: {$exists: true},                                                                      // 417
      expires: {$exists: false},                                                                             // 418
    }, {fields: {_id: 1}}).forEach(function (user) {                                                         // 419
      sendReferralProgramNotification(user._id);                                                             // 420
    });                                                                                                      // 421
  }                                                                                                          // 422
}                                                                                                            // 423
                                                                                                             // 424
// This must come after all the functions named within are defined.                                          // 425
// Only append to this list!  Do not modify or remove list entries;                                          // 426
// doing so is likely change the meaning and semantics of user databases.                                    // 427
var MIGRATIONS = [                                                                                           // 428
  updateLoginStyleToRedirect,                                                                                // 429
  enableLegacyOAuthProvidersIfNotInSettings,                                                                 // 430
  denormalizeInviteInfo,                                                                                     // 431
  mergeRoleAssignmentsIntoApiTokens,                                                                         // 432
  fixOasisStorageUsageStats,                                                                                 // 433
  fetchProfilePictures,                                                                                      // 434
  assignPlans,                                                                                               // 435
  removeKeyrings,                                                                                            // 436
  useLocalizedTextInUserActions,                                                                             // 437
  verifyAllPgpSignatures,                                                                                    // 438
  splitUserIdsIntoAccountIdsAndIdentityIds,                                                                  // 439
  appUpdateSettings,                                                                                         // 440
  moveDevAndEmailLoginDataIntoIdentities,                                                                    // 441
  repairEmailIdentityIds,                                                                                    // 442
  splitAccountUsersAndIdentityUsers,                                                                         // 443
  populateContactsFromApiTokens,                                                                             // 444
  cleanUpApiTokens,                                                                                          // 445
  initServerTitleAndReturnAddress,                                                                           // 446
  sendReferralNotifications,                                                                                 // 447
];                                                                                                           // 448
                                                                                                             // 449
function migrateToLatest() {                                                                                 // 450
  if (Meteor.settings.replicaNumber) {                                                                       // 451
    // This is a replica. Wait for the first replica to perform migrations.                                  // 452
                                                                                                             // 453
    console.log("Waiting for migrations on replica zero...");                                                // 454
                                                                                                             // 455
    var done = new Future();                                                                                 // 456
    var change = function (doc) {                                                                            // 457
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                   // 458
      if (doc.value >= MIGRATIONS.length) done.return();                                                     // 459
    }                                                                                                        // 460
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                                    // 461
      added: change,                                                                                         // 462
      changed: change                                                                                        // 463
    });                                                                                                      // 464
                                                                                                             // 465
    done.wait();                                                                                             // 466
    observer.stop();                                                                                         // 467
    console.log("Migrations have completed on replica zero.");                                               // 468
                                                                                                             // 469
  } else {                                                                                                   // 470
    var applied = Migrations.findOne({_id: "migrations_applied"});                                           // 471
    var start;                                                                                               // 472
    if (!applied) {                                                                                          // 473
      // Migrations table is not yet seeded with a value.  This means it has                                 // 474
      // applied 0 migrations.  Persist this.                                                                // 475
      Migrations.insert({_id: "migrations_applied", value: 0});                                              // 476
      start = 0;                                                                                             // 477
    } else {                                                                                                 // 478
      start = applied.value;                                                                                 // 479
    }                                                                                                        // 480
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                           // 481
                                                                                                             // 482
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                                      // 483
      // Apply migration i, then record that migration i was successfully run.                               // 484
      console.log("Applying migration " + (i+1));                                                            // 485
      MIGRATIONS[i]();                                                                                       // 486
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                                  // 487
      console.log("Applied migration " + (i+1));                                                             // 488
    }                                                                                                        // 489
  }                                                                                                          // 490
}                                                                                                            // 491
                                                                                                             // 492
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                     // 493
                                                                                                             // 494
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
